// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;
import de.uka.ilkd.key.logic.*;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * An InitialPositionTable is a PositionTable that describes the
 * beginning of the element/subelement relationship. Thus, an
 * InitialPositionTable describes the information on where the
 * semisequents of a sequent are located. It is the root of the tree of
 * PositionTables and may be asked for a PosInSequent for a given index
 * position and a given Sequent.
 */
public class InitialPositionTable extends PositionTable{

    private int start = 1;
    private int end   = 1;

    /**
     * creates a new Initial PositionTable with the number of  
     * Semisequents in a
     * Sequent. So the rows should be 2.
     * @param rows the number of direct sub-elements (here: semisequents) in
     * the sequent whose position information is represented by the
     * constructed object. 
     * @param start sets the width of chars before the beginning of the sequent
     * @param end sets the width of chars after teh end of the sequent
     */
    public InitialPositionTable(int rows, int start, int end) {
	super(rows);
	this.start = start;
	this.end = end;
    }

    /**
     * returns the PosInSequent for a given char position and the given Sequent
     * @param index the char position that points to the wanted position in 
     * sequent
     * @param seq the sequent from where to retrieve the position in sequent
     */
    public PosInSequent getPosInSequent(int index, Sequent seq){
	//	if (index < startPos[0]+start || index > endPos[1] - end) {
	//  return null;
	//}
	LinkedList posList= getPositionList(index, new LinkedList());
	int posListSize=posList.size();
	int[] posArray=new int[posListSize];
	int k=0;
	for (Iterator it=posList.iterator(); it.hasNext(); k++) {
	    posArray[k]=((Integer)it.next()).intValue();
	}

	if (posListSize == 0) {	    	    
	    if ( index >= startPos[0] && index <= endPos[1]) {
		PosInSequent posInSeq=PosInSequent.createSequentPos();	
		posInSeq.setBounds(startPos[0], endPos[1]);
		return posInSeq;       
	    } else {
		return null;
	    }
	}
	int offset=0;

	//depth>=1, so this is at least a semisequent
	int semiseqNo = posArray[0];
	Semisequent semiseq=(semiseqNo == 0) ? seq.antecedent() : seq.succedent();
	PositionTable semiseqPosTable=child[semiseqNo];
	if (posListSize == 1) {
	    PosInSequent posInSeq = (semiseqNo == 0) 
		? PosInSequent.createAntecedentPos()
		: PosInSequent.createSuccedentPos();
	    posInSeq.setBounds(startPos[semiseqNo], endPos[semiseqNo]);
	    return posInSeq;
	}
	offset = offset+startPos[semiseqNo];

	//depth >=2, so we have at least a ConstrainedFormula
	int cfmaNo = posArray[1];
	ConstrainedFormula cfma = semiseq.get(cfmaNo);
	PositionTable cfmaTable = semiseqPosTable.child[cfmaNo];
	if (posListSize == 2) {
	    PosInSequent posInSeq = PosInSequent.createCfmaPos
		(new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, seq), 
		 false, true);
	    posInSeq.setBounds(offset+semiseqPosTable.startPos[cfmaNo], 
			       offset+semiseqPosTable.endPos[cfmaNo]);
	    return posInSeq;
	}
	offset = offset+semiseqPosTable.startPos[cfmaNo];

	//depth>=3, so we have at least a constraint or a formula
	int fmaOrConstrNo = posArray[2];
	PositionTable fmaOrConstrTable = cfmaTable.child[fmaOrConstrNo];	
	if (posListSize == 3) {
	    PosInSequent posInSeq = PosInSequent.createCfmaPos
		(new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, seq), 
		 (fmaOrConstrNo == 1), false);
	    posInSeq.setBounds(offset+cfmaTable.startPos[fmaOrConstrNo],
			       offset+cfmaTable.endPos[fmaOrConstrNo]);
	    return posInSeq;
	}
	offset = offset+cfmaTable.startPos[fmaOrConstrNo];

	//depth > = 4, so we are somewhere in a Term
	
	//is a JavaBlock meant?
	if (cfmaTable.child[0].child[posArray[3]] instanceof ProgramPositionTable) {
	    PosInSequent posInSeq = PosInSequent.createJavaBlockPos
		(new PosInOccurrence(cfma, PosInTerm.TOP_LEVEL, seq));
	    posInSeq.setBounds(offset+cfmaTable.child[0].startPos[posArray[3]],
			       offset+cfmaTable.child[0].endPos[posArray[3]]);

	    ProgramPositionTable ppt = (ProgramPositionTable)cfmaTable.child[0].child[posArray[3]];
	    posInSeq.setFirstJavaStatementRange(ppt.getFirstStatementRange());
	    return posInSeq;
	    
	}

	
	//is first "child" of a top level term a JavaBlock?

	PositionTable currentTable = fmaOrConstrTable;
	PosInTerm currentPosInTerm = PosInTerm.TOP_LEVEL;
	for (int i = 3; i<posListSize; i++) {
	    int subNo  =  posArray[i];
	    if (subNo != 0 && currentTable.child[0] instanceof ProgramPositionTable) {
		currentPosInTerm = currentPosInTerm.down(subNo-1);	       
	    } else if (subNo == 0 && 
		       currentTable.child[0] instanceof ProgramPositionTable) {
		// we are inside a java block no further navigation at this time
		// has to be added later
		PosInSequent posInSeq = PosInSequent.createJavaBlockPos
		    (new PosInOccurrence(cfma, currentPosInTerm, seq));	       
		posInSeq.setBounds(offset+currentTable.startPos[subNo], 
				   offset+currentTable.endPos[subNo]);

		ProgramPositionTable ppt = (ProgramPositionTable)currentTable.child[0];
		posInSeq.setFirstJavaStatementRange(ppt.getFirstStatementRange());

		return posInSeq;				
	    } else {
		currentPosInTerm = currentPosInTerm.down(subNo);
	    }
	    if (posListSize == i+1) {
		PosInSequent posInSeq  =  PosInSequent.createCfmaPos
		    (new PosInOccurrence(cfma, currentPosInTerm, seq), 
		     false, false);	       
		posInSeq.setBounds(offset+currentTable.startPos[subNo], 
				   offset+currentTable.endPos[subNo]);
		return posInSeq;		
	    }
	    offset = offset+currentTable.startPos[subNo];
	    currentTable = currentTable.child[subNo];

	}	
	return null; // never reached
    }

}


