// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ProblemLoader;
import de.uka.ilkd.key.proof.ProofSaver;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYResourceManager;
import java.net.URL;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;

// just for loading remove if load moves to another class 
import java.io.*;

public class Main extends JFrame implements java.io.Serializable {

//DO NOT CHANGE THE FOLLOWING 2 LINES, UNLESS YOU KNOW ABOUT PRCS-REPLACEMENT
    /* $Format: "    private static final String PRCSVERSION = \"$ProjectVersion$\";"$ */
    private static final String PRCSVERSION = "0.475";

    private static final String VERSION="0.8 (internal: "+PRCSVERSION+")";
    private static final String COPYRIGHT="(C) Copyright 2001-2003 Universit\u00e4t"
	+" Karlsruhe and Chalmers University of Technology";

    /** the tab bar at the left */
    private JTabbedPane	tabbedPane;
       
    /** the current goal view */
    private JScrollPane goalView;

    /** the current proof tree*/
    private JScrollPane proofView;

    /** the list of current open goals*/
    private JScrollPane openGoalsView;

    /** the view of a sequent */
    private SequentView sequentView;

    /** the user constraint view */
    private UserConstraintView userConstraintView = null;

    /** the mediator is stored here */
    protected KeYMediator mediator;
    
    /** the status line (see inner class) */
    StatusLine statusLine;
    
    /** keeps track of last opened directory (for load)
     */
    protected File directory = new File(".");

    
    /** listener to global proof events */
    private MainProofListener proofListener;

    /** listener to gui events */
    private MainGUIListener guiListener;

    protected JMenu file = new JMenu("File");
    protected JMenu view = new JMenu("View");
    protected JMenu options = new JMenu("Options");
    protected JMenu help = new JMenu("Help");
    protected JMenuBar menubar;
    
    private final String PARA = 
       "<p style=\"font-family: lucida;font-size: 12pt;font-weight: bold\">";
    
//    private JButton heuristicsButton = new JButton("Activate Heuristics");
    private JCheckBox resumeHeuristics = 
       new JCheckBox("<html>"+PARA+"Autoresume"+PARA+"heuristics");
    private JButton applyHeuristicsButton = new JButton("Apply Heuristics");

    /** the proofsettings */
    private ProofSettings proofSettings;

    public Main(){
	super();
    }

    /** creates main frame 
     * @param title a String the frames title
     * @param proofSettings the settings the proof uses (such like active
     * heuristics, LDTs etc.) 
     */ 
    public Main(String title, ProofSettings settings) {
	super(title);
	setIconImage(IconFactory.keyLogo());
	setDefaultCloseOperation(WindowConstants.
				 DO_NOTHING_ON_CLOSE);
	proofSettings     = settings;	
	proofListener     = new MainProofListener();
	guiListener       = new MainGUIListener();
	layoutMain();
	setSize(700,500);
	addWindowListener(new MainListener());
    }

    /** returns the proof settings 
     * @return the proof settings 
     */
    public ProofSettings proofSettings() {
	return proofSettings;
    }

    public void updateUI() {
	if (goalView != null) {
	    goalView.updateUI();
	} 
	if (proofView != null) {
	    proofView.updateUI();
	}
	if (openGoalsView != null) {
	    openGoalsView.updateUI();
	}
	if (userConstraintView != null) {
	    userConstraintView.updateUI();
	}

    }

    /** sets the mediator to corresspond with other gui elements
     * @param m the KeYMediator
     */
    public void setMediator(KeYMediator m) {
	if (mediator !=  null) {
	    unregister();
	}
	mediator = m;
	sequentView = new SequentView(mediator);
	if ( userConstraintView != null )
	    userConstraintView.setMediator ( m );
	register();
    }

    /** register several listeners */
    private void register() {	
	mediator().addKeYSelectionListener(proofListener);
	mediator().addInteractiveProofListener(proofListener);
	mediator().addGUIListener(guiListener);	
    }

    /** unregister several listeners */
    private void unregister() {
	mediator().removeKeYSelectionListener(proofListener);
	mediator().removeInteractiveProofListener(proofListener);
	mediator().removeGUIListener(guiListener);	
    }
   

    /** return the mediator
     * @return the mediator
     */
    public KeYMediator mediator() {
	return mediator;
    }

    /** paints empty view */
    private void paintEmptyViewComponent(JScrollPane pane, 
					 String name) { 	
	pane.setBorder(new TitledBorder(name));
	pane.setBackground(Color.white);
	pane.getViewport().setBackground(Color.white);	
	pane.setMinimumSize(new Dimension(150,100));
    }

    /** adds a new tab */
    public void addTab(String tabname, JScrollPane view,
		       String description) {
	tabbedPane.addTab(tabname, null, view, description);
    }

    /** lays out the main frame */
    protected void layoutMain() {
	setJMenuBar(createMenuBar());
	getContentPane().setLayout(new BorderLayout());	

	goalView = new JScrollPane();	
	paintEmptyViewComponent(goalView, "Current Goal");

	proofView = new JScrollPane();
	paintEmptyViewComponent(proofView, "Proof");

	openGoalsView = new JScrollPane();
	paintEmptyViewComponent(openGoalsView, "Open Goals");

	userConstraintView = new UserConstraintView ();

	tabbedPane = new JTabbedPane();

	addTab("Proof", proofView, 
	       "The current state of the "
	       +"proof as tree");

	addTab("Goals", openGoalsView, 
	       "The currently open goals");

	/* Constraint view is currently disabled
	 *
	   tabbedPane.addTab("User Constraint", null, userConstraintView, 
	   "Currently choosen metavariable instantiations"); */

	tabbedPane.setSelectedIndex(0);

	JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
					      tabbedPane, goalView);
	splitPane.setResizeWeight(0); // the right pane is more important
	getContentPane().add(splitPane, BorderLayout.CENTER);
	
	statusLine = new StatusLine(true);
        getContentPane().add(statusLine, BorderLayout.SOUTH);
        
        JPanel buttonPanel = new JPanel();
	buttonPanel.add(applyHeuristicsButton);
//	buttonPanel.add(heuristicsButton);
	buttonPanel.add(resumeHeuristics);
	buttonPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
	getContentPane().add(buttonPanel, BorderLayout.NORTH);
//        Dimension dim = heuristicsButton.getPreferredSize();
//        heuristicsButton.setPreferredSize(
//           new Dimension(dim.width+20, dim.height));
//        dim = applyHeuristicsButton.getPreferredSize();
//        applyHeuristicsButton.setPreferredSize(
//           new Dimension(dim.width+20, dim.height));
        resumeHeuristics.addItemListener(new ItemListener() {
           public void itemStateChanged(ItemEvent e) {
              if (e.getStateChange() == ItemEvent.SELECTED) {
                 mediator().setAutoHeuristics(true);
              } 
              else
              if (e.getStateChange() == ItemEvent.DESELECTED) {
                 mediator().setAutoHeuristics(false);
              }
              else System.err.println("Heuristics checkbox undefined state: "+
                                      e.getStateChange());
              
           }
        });
        applyHeuristicsButton.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
		      mediator().startHeuristics();
	   }
        });	
        
        JButton goalBackButton = new JButton("Goal Back");
	buttonPanel.add(goalBackButton);
        goalBackButton.addActionListener(new ActionListener() {
	   public void actionPerformed(ActionEvent e) {
	       setBack();
	   }
        });


    }

    public void setStandardStatusLine() {
       setStatusLine("Integrated Deductive Software Design: Ready");
    }


    public void setStatusLine(String s, int totalChars) {
       final String str = s;
       final int max = totalChars;
       Runnable lineUpdater = new Runnable() {
          public void run() {
             statusLine.setStatusLine(str, max);
	     statusLine.addProgressBar();
          }
       };
       try{
          SwingUtilities.invokeAndWait(lineUpdater);
       } catch(Exception e) {
          System.err.println(e);
          e.printStackTrace();
       }
    }

    public void setStatusLine(String s) {
       final String str = s;
       Runnable lineUpdater = new Runnable() {
          public void run() {
             statusLine.setStatusLine(str);
	     statusLine.removeProgressBar();
          }
       };
       try{
          SwingUtilities.invokeAndWait(lineUpdater);
       } catch(Exception e) {
          System.err.println(e);
          e.printStackTrace();
       }
    }

    public void setProgress(int value){
	statusLine.setProgress(value);
    }

    /** ensure settings have been saved */
    protected void settingsSaved() {
	if (proofSettings().changed()) {
	    int option = JOptionPane.showConfirmDialog
		(Main.this, "Settings have changed. Save them?", 
		 "Save Settings", 
		 JOptionPane.YES_NO_OPTION);
	    if (option == JOptionPane.YES_OPTION) {
		proofSettings().saveSettings();
	    }
	}
    }

    /** exit  */
    protected void exitMain() {
	settingsSaved(); 
	int option = JOptionPane.showConfirmDialog
	    (Main.this, "Really Quit?", "Exit", 
	     JOptionPane.YES_NO_OPTION);
	if (option == JOptionPane.YES_OPTION) {
	    System.out.println("Have a nice day.");
	    System.exit(-1);
	}
    }

    /** opens selection dialog for heuristics */
    protected void selectHeuristics() {
	HeuristicSelector selector = new HeuristicSelector
	    (mediator(), mediator().heur_ns(), 
	     proofSettings().getHeuristicSettings());
    }

    /** opens selection dialog for LDT models */
    protected void selectLDT() {
	LDTSelector selector = new LDTSelector
	    (mediator(), 
	     proofSettings().getLDTSettings());
    }

    protected void setBack() {
       mediator.setBack();
    }

    protected void makePrettyView() {

	if (mediator().ensureProofLoaded()) {
	    PresentationFeatures.ENABLED=!PresentationFeatures.ENABLED;
	    if (!PresentationFeatures.ENABLED) {
		mediator().getNotationInfo().setBackToDefault();
	    } else {	    
		PresentationFeatures.modifyNotationInfo(mediator.getNotationInfo(),
							mediator.func_ns());
	    }
	    mediator().getSelectedProof().updateProof();	
	}

    }

    protected void showLicense() {
			     
	URL lic = 
	KeYResourceManager.getManager().getResourceFile(Main.class,
							"LICENSE.TXT"); 
	StringBuffer sb=new StringBuffer();
	try {
	    FileInputStream inp=new FileInputStream(lic.getFile());
	    while (inp.available()>0) sb.append((char)inp.read());	   
	} catch (Exception e) {
	    sb=new StringBuffer(COPYRIGHT+"\nKeY is protected by the "
				+"GNU General Public License");
	}
	String s=sb.toString();
	JScrollPane scroll = new JScrollPane();
	JTextArea text = new JTextArea(s,20,40);
	text.setEditable(false);
	text.setCaretPosition(0);
	scroll.setViewportView(text);
	JFrame fr = new JFrame("KeY License");
	fr.getContentPane().setLayout(new BorderLayout());
	fr.getContentPane().add(scroll,BorderLayout.CENTER);
	JButton ok = new JButton("OK");
	ok.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    //		    ((JFrame)((JRootPane)((JLayeredPane)((JPanel)((JButton)e.getSource()).getParent()).getParent()).getParent()).getParent()).dispose();
		    ((JFrame)((JButton)e.getSource())
		     .getTopLevelAncestor()).dispose();
		}});
	fr.getContentPane().add(ok, BorderLayout.SOUTH);
	fr.setSize(600,900);
	fr.getContentPane().add(scroll);
	fr.show();
    }



    protected void showAbout() {
	JOptionPane pane
	    =new JOptionPane(COPYRIGHT+"\n\n"
			     +"WWW: http://i12www.ira.uka.de/~key\n\n"
			     +"Version "+VERSION,
			     JOptionPane.INFORMATION_MESSAGE,
			     JOptionPane.DEFAULT_OPTION,
			     IconFactory.keyLogo(105,60));
	JDialog dialog = pane.createDialog(this, "The KeY Project");
	dialog.show();
    }

    /** registers a new menuitem at the given menu
     * @param menu the JMenu where to register the new item
     * @param item the JMenuItem to register
     */
    public void registerAtMenu(JMenu menu, JMenuItem item) {
	menu.add(item);
    }

    /** registers a new menuitem at the file menu
     * @param item the JMenuItem to register
     */
    public void registerAtFileMenu(JMenuItem item) {	
	registerAtMenu(file, item);
    }

    /** registers a new menuitem at the view menu
     * @param item the JMenuItem to register
     */    
    public void registerAtViewMenu(JMenuItem item) {
	registerAtMenu(view, item);
    }

    /** registers a new menuitem at the options menu
     * @param item the JMenuItem to register
     */
    public void registerAtOptionsMenu(JMenuItem item) {	
	registerAtMenu(options, item);
    }
    
    /** registers a new menuitem at the help menu
     * @param item the JMenuItem to register
     */
    public void registerAtHelpMenu(JMenuItem item) {
	registerAtMenu(help, item);
    }

    /** adds a separator to the given menu
     */
    public void addSeparator(JMenu menu) {	
	menu.addSeparator();
    }


    /** starts the goal choice frame */
    protected void initGoalList() {
	GoalList goalList = new GoalList(mediator);
	goalList.setSize(goalList.getPreferredSize());
	openGoalsView.setViewportView(goalList);
    } 

    /** starts the gui proof tree frame */
    protected void initGUIProofTree() {
	ProofTreeView guiProofTree = new ProofTreeView(mediator());
	guiProofTree.setSize(guiProofTree.getPreferredSize());
	guiProofTree.setVisible(true);
	proofView.setViewportView(guiProofTree);
    } 

    protected void createFileMenu() {
	file = new JMenu("File");
	JMenuItem load = new JMenuItem("Load");
	load.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_O, ActionEvent.CTRL_MASK));

	load.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JFileChooser jFC = new JFileChooser(directory);
		    int loaded = jFC.showOpenDialog(Main.this);
		    if (loaded == JFileChooser.APPROVE_OPTION) {
			File file = jFC.getSelectedFile(); 
			loadProblem(file);
		    }
		}
	    });

	JMenuItem save = new JMenuItem("Save");
	save.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_S, ActionEvent.CTRL_MASK));
	save.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    saveProof();
		}});


	JMenuItem exit = new JMenuItem("Exit");
	exit.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
	exit.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    exitMain();		    
		}});	

	registerAtFileMenu(load);
	registerAtFileMenu(save);
	addSeparator(file);
	registerAtFileMenu(exit);
    }


    protected void createViewMenu() {
	view = new JMenu("View");

	JMenuItem pretty = new JCheckBoxMenuItem("Pretty&Untrue");
	pretty.setSelected(PresentationFeatures.ENABLED);
	pretty.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    makePrettyView();
		}});

	final JMenuItem smaller = new JMenuItem("Smaller");
	smaller.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    Config.DEFAULT.smaller();
		}});
	final JMenuItem larger  = new JMenuItem("Larger");
	larger.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    Config.DEFAULT.larger();
		}});

	Config.DEFAULT.addConfigChangeListener(
	    new ConfigChangeListener() {
		    public void configChanged(ConfigChangeEvent e) {
			smaller.setEnabled(!Config.DEFAULT.isMinimumSize());
			larger.setEnabled(!Config.DEFAULT.isMaximumSize());
		    }
		});
	
	registerAtViewMenu(pretty);
	addSeparator(view);
	registerAtViewMenu(smaller);
	registerAtViewMenu(larger);
    }


    protected void createOptionsMenu() {
	options = new JMenu("Options");
	JMenuItem heuristicItem = new JMenuItem("Heuristics...");
	heuristicItem.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_H, ActionEvent.CTRL_MASK));
	heuristicItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    selectHeuristics();
		}});
	registerAtOptionsMenu(heuristicItem);

	JMenuItem dataTypeItem = new JMenuItem("LDT Models...");
	dataTypeItem.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_L, ActionEvent.CTRL_MASK));

	dataTypeItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    selectLDT();
		}});
	registerAtOptionsMenu(dataTypeItem);	
	addSeparator(options);

	proofSettings().item().addActionListener
	    (new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			proofSettings().saveSettings();
		    }});

	registerAtOptionsMenu(proofSettings().item());

    }


    protected void createHelpMenu() {
	help = new JMenu("Help");
	JMenuItem about = new JMenuItem("About");
	about.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    showAbout();
		}});
	help.add(about);	

	JMenuItem license = new JMenuItem("License");
	license.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    showLicense();
		}});
	help.add(license);

    }

    /** creates menubar */
    protected JMenuBar createMenuBar() {
	menubar = new JMenuBar();
	createFileMenu();
	createViewMenu();
	createOptionsMenu();
	createHelpMenu();

	menubar.add(file);
	menubar.add(view);
	menubar.add(options);
        menubar.add(Box.createHorizontalGlue());
	menubar.add(help);
	
	return menubar;
    }


    /** updates the view of the sequent being displayed in the main
     * frame 
     */
    public synchronized void updateGoalView(Sequent sequent) {
	SequentPrinter printer = new SequentPrinter
	    (new ProgramPrinter(null), 
	     mediator().getNotationInfo(),
	     sequent);
	sequentView.setPrinter(printer);
	sequentView.printSequent();

	goalView.setViewportView(sequentView);
	goalView.setBorder(new TitledBorder("Current Goal"));
        sequentView.setCaretPosition(0); // at least until a better solution
    }

    /** saves a proof */
    protected void saveProof() {
	JFileChooser jFC = new JFileChooser(directory);
	int saved = jFC.showSaveDialog(this);
	if (saved == JFileChooser.APPROVE_OPTION) {
	    String filename = jFC.getSelectedFile().getName();    
	    filename = (filename.endsWith(".proof") ? filename : 
			filename+".proof");
            ProofSaver saver = new ProofSaver(mediator(), filename);
            String errorMsg = saver.save();
            if (errorMsg != null) {
	       JOptionPane.showMessageDialog(this, errorMsg, "Oops...", 
                                             JOptionPane.ERROR_MESSAGE);
            }

	}	
    }


    protected void loadProblem(File file) {
       Thread t = new Thread(new ProblemLoader(file, this));
       t.start();
    }


    public void addProblem(String problemID, Term problem, TacletIndex rules,
			   Namespace variables, Namespace functions, 
			   Namespace sorts, Namespace heuristics,
			   JavaInfo javaInfo) {

        setUpNewProof(problemID, problem, rules, 
		      variables, functions, sorts, 
		      heuristics, javaInfo);

        initGoalList();
        initGUIProofTree();
        SwingUtilities.updateComponentTreeUI(this);
        PresentationFeatures.initialize(functions, 
                                        mediator.getNotationInfo(),
				        mediator.getSelectedProof());
	setStandardStatusLine();
    }


    protected Proof setUpNewProof(String problemID,
				  Term problem,
				  TacletIndex rules,
				  Namespace variables,
				  Namespace functions,
				  Namespace sorts,
				  Namespace heuristics,
				  JavaInfo javaInfo) {

	ConstrainedFormula cf = new ConstrainedFormula(problem);
	Proof proof = new Proof(problemID);
	proof.setUp(
           new Node(proof, Sequent.createSuccSequent(
              Semisequent.EMPTY_SEMISEQUENT.insert(0,cf))), 
           rules);
	proof.openGoals().head().addProgramVariables(javaInfo.programVariables());
        
	KeYMediator mediator = mediator();
	mediator.setProof(proof);
	mediator.setNamespaces(variables, functions, sorts, heuristics);
	mediator.setJavaInfo(javaInfo);
        mediator.setupHeuristics();
	return proof;
    }

    

    class MainListener extends WindowAdapter 
                       implements java.io.Serializable {


	public void windowClosing(WindowEvent e) {
	    exitMain();	    
	}
    }

    class MainGUIListener implements GUIListener, java.io.Serializable {
	/** invoked if a frame that wants modal access is opened */

	private void enableMenuBar(JMenuBar  m, boolean b){
	    for (int i=0; i < m.getMenuCount(); i++){
		JMenu currentMenu = m.getMenu(i);
		if (currentMenu instanceof JMenu) 
                   currentMenu.setEnabled(b); // otherwise it's a spacer
	    }
	}

	public void modalDialogOpened(GUIEvent e) {
	 
	    if (e.getSource() instanceof ApplyTacletDialog){
		// disable all elements except the sequent window (drag'n'drop !) ...
		enableMenuBar(Main.this.getJMenuBar(), false); 
		Main.this.goalView.setEnabled(false);
		Main.this.proofView.setEnabled(false); 
		Main.this.openGoalsView.setEnabled(false);
//FIXEM: add button row
	    } else {
		// disable the whole main window ...
		Main.this.setEnabled(false);
	    }
	}

	/** invoked if a frame that wants modal access is closed */
	public void modalDialogClosed(GUIEvent e) {
	    if (e.getSource() instanceof ApplyTacletDialog){
		// enable all previously diabled elements ...
		enableMenuBar(Main.this.getJMenuBar(), true); 
		Main.this.goalView.setEnabled(true);
		Main.this.proofView.setEnabled(true); 
		Main.this.openGoalsView.setEnabled(true);
	    } else {
		// enable the whole main window ...
	       Main.this.setEnabled(true);
	    }
	}
	
	public void shutDown(GUIEvent e) {
	    Main.this.dispose();
	}

    }

    class MainProofListener implements InteractiveProofListener, 
                                       KeYSelectionListener,
	                               java.io.Serializable {

	/** set to true if the view of the current goal
	 * should not be updated */       
	private boolean disableCurrentGoalView = false;

	private synchronized void setProofNodeDisplay() {
 	    if (!disableCurrentGoalView) {
                Goal goal;
                try {
		   goal = mediator().getSelectedGoal();
                } catch(IllegalStateException e) { // there is no proof (yet)
                   return;
                }
		if (goal != null) {
		    updateGoalView(goal.seq());
		} else {
		    NonGoalInfoView innerNodeView = 
 			new NonGoalInfoView(mediator().getSelectedNode(), 
 					    mediator());
		    goalView.setBorder(new TitledBorder("Inner Node"));
		    goalView.setViewportView(innerNodeView);	    
		    goalView.setViewportView(innerNodeView);
                    innerNodeView.setCaretPosition(0);
		}
		goalView.validate();	    
		validate();
	    }
	}

	/** focused node has changed */
 	public synchronized void selectedNodeChanged(KeYSelectionEvent e) {
	    setProofNodeDisplay();	    
 	}

 	/** the selected proof has changed (e.g. a new proof has been
 	 * loaded) */ 
 	public synchronized void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("Main: initialize with new proof");
 	    disableCurrentGoalView = false;	    
 	    goalView.setViewportView(null);		
	    setProofNodeDisplay();
 	}


	/** invoked if automatic execution has started
	 */
	public synchronized void autoModeStarted(ProofEvent e) {
	    disableCurrentGoalView = true;
	    mediator().removeKeYSelectionListener(proofListener);
	}

	/** invoked if automatic execution has stopped
	 */
	public synchronized void autoModeStopped(ProofEvent e) {
	    disableCurrentGoalView = false;
	    setProofNodeDisplay();
	    mediator().addKeYSelectionListener(proofListener);
	}

	/** invoked when a rule has been applied */
	public synchronized void ruleApplied(ProofEvent e) {
	    //	    setProofNodeDisplay();
	}
    }

    public static void evaluateOptions(String[] opt) {
	// all strings to upper case	
	for (int i = 0; i < opt.length; i++) {
	    opt[i] = opt[i].toUpperCase();
	}

	int index = 0;
	while (opt.length > index) {	    
	    if (opt[index].equals("NO_DEBUG")) {
		de.uka.ilkd.key.util.Debug.ENABLE_DEBUG = false;
	    } else if (opt[index].equals("DEBUG")) {
		de.uka.ilkd.key.util.Debug.ENABLE_DEBUG = true;
	    } else if (opt[index].equals("NO_ASSERTION")) {
		de.uka.ilkd.key.util.Debug.ENABLE_ASSERTION = false;
	    } else if (opt[index].equals("ASSERTION")) {
		de.uka.ilkd.key.util.Debug.ENABLE_ASSERTION = true;
	    } else {
		System.out.println("Possible parameters are (* = default): ");
		System.out.println("  no_debug        : disables debug mode");
		System.out.println("  debug           : enables debug mode (*)");
		System.out.println("  no_assertion    : disables assertions");
		System.out.println("  assertion       : enables assertions (*)");
		System.exit(-1);
	    }
	    index++;
	}	

	if (Debug.ENABLE_DEBUG) {
	    System.out.println("Running in debug mode ...");
	} else {
	    System.out.println("Running in normal mode ...");
	}
	if (Debug.ENABLE_ASSERTION) {
	    System.out.println("Using assertions ...");	   
	} else {
	    System.out.println("Not using assertions ...");	   
	}

    }
    
    
    class StatusLine extends JPanel {
	private JLabel text;
	private JProgressBar progressBar;
	private int max=0;
    
       StatusLine(boolean withProgressBar) {
           super();
	   this.setBorder(new BevelBorder(BevelBorder.LOWERED));
	   this.setBackground(Color.gray);
	   this.setFont(Main.this.getFont());
	   this.setOpaque(false);

           text = new JLabel("<html>"+PARA+COPYRIGHT+
              PARA+
              "KeY is free software and comes with ABSOLUTELY NO WARRANTY."+
	      " See Help | License.");

           text.setIcon(IconFactory.keyLogo(35,20));
           this.add(text);

	   if (withProgressBar) {
	       progressBar = new JProgressBar(0, max);
	       progressBar.setValue(0);
	       progressBar.setStringPainted(true);
	       this.add(progressBar);
	   }
	   this.setVisible(true);
       }

	public void setMaximum(int value){
	    progressBar.setMaximum(value);
	}

	public void setProgress(int value){
	    final int v=value;
	    Runnable doSetProgressBarValue = new Runnable() {
		    public void run() {
			progressBar.setValue(v);
		    }
		};
	    SwingUtilities.invokeLater(doSetProgressBarValue);
	}

	public void addProgressBar(){
	    this.add(progressBar);
	}

	public void removeProgressBar(){
	    this.remove(progressBar);
	}

       public void setStatusLine(String s, int totalChars) {
           text.setText(s);
	   setMaximum(totalChars);
       }

       public void setStatusLine(String s) {
           text.setText(s);
       }
    }    
    

    
    public static void main(String[] args) {	
	Main.evaluateOptions(args);
	Config.DEFAULT.setDefaultFonts();
	Main key = new Main("KeY -- Prover", ProofSettings.THE_SETTINGS);
	KeYMediator mediator = 
	    new KeYMediator(key);
	key.setMediator(mediator);
	key.initGoalList();
	key.initGUIProofTree();
	SwingUtilities.updateComponentTreeUI(key);
	key.show();
    }
    
    
}
