// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java;

import java.util.HashMap;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;

public class TypeConverter {


    /** counts number of created object used to create unique names. 
     * %%%% HACK! 
     */
    private static long OBJECT_COUNTER = 0;
    
    private Services services;

    private TermFactory termFab = TermFactory.DEFAULT;

    private LDT integerLDT;
    private LDT booleanLDT;
    private LDT objectLDT;
    private LDT[] models = new LDT[0];


    public TypeConverter(Services s){
        services = s;
    }

    public void init(LDT integerLDT, LDT booleanLDT, LDT objectLDT) {
	this.integerLDT = integerLDT;
	this.booleanLDT = booleanLDT;
	this.objectLDT = objectLDT;
	this.models = new LDT[]{integerLDT, booleanLDT, objectLDT};
    }

    public LDT[] getModels() {
	return models;
    }

    public LDT getModelFor(Type t) {
	if (integerLDT == null || booleanLDT == null || objectLDT == null) {
	    return  null;
	}
	
	if (integerLDT.javaType().contains(t)) {
	    return integerLDT;
	} else if (booleanLDT.javaType().contains(t)) {
	    return booleanLDT;
	} else if (objectLDT.javaType().contains(t)) {
	    return objectLDT;
	}
	Debug.out("typeconverter: No LDT found for ", t);
	return null;
    }

    public LDT getModelFor(Sort s) {
	if (integerLDT == null || booleanLDT == null || objectLDT == null) {
	    return  null;
	}
	if (integerLDT.targetSort() == s) {
	    return integerLDT;
	} else if (booleanLDT.targetSort() == s) {	    
	    return booleanLDT;
	} else if (objectLDT.targetSort() == s) {
	    return booleanLDT;
	}
	Debug.out("No LDT found for ", s);
	return null;
    }

    public NumberLDT getIntegerLDT() {
	return (NumberLDT)integerLDT;
    }

    public Sort getObjectSort() {
	if (objectLDT==null) return null;
	return objectLDT.targetSort();
    }


    public KeYJavaType getObjectKeYJavaType() {
	return objectLDT.getKeYJavaType(objectLDT.javaType().head());
    }

		     
    private Term translateOperator(de.uka.ilkd.key.java.expression.Operator op,
				   LDT intLDT, LDT booleanLDT, LDT objectLDT) {
	Term[] subs  = new Term[op.getArity()];
	if (op.getArity() >= 1) {
	    subs[0] = convertToLogicElement(op.getExpressionAt(0));
	} 
	if (op.getArity() == 2) {
	    subs[1] = convertToLogicElement(op.getExpressionAt(1));	  
	}	
	
	LDT responsibleLDT = null;
	if (intLDT.isResponsible(op, subs)) {
	    responsibleLDT = intLDT;
	} else if (booleanLDT.isResponsible(op, subs)) {
	    responsibleLDT = booleanLDT;
	} else if (objectLDT.isResponsible(op, subs)) {
	    responsibleLDT = objectLDT;
	} else{
	    Debug.out("typeconverter: no data type model "+
		      "available to convert:", op, op.getClass());		
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	}
	return termFab.createFunctionTerm
	    (responsibleLDT.getFunctionFor(op), subs);
    }


    private Term convertReferencePrefix(ReferencePrefix prefix) {
	Debug.out("typeconverter: (prefix, class)", prefix, 
		  (prefix != null ? prefix.getClass() : null));		
	if (prefix instanceof EmbededVariableReference) {
	    return termFab.createVariableTerm
		(((EmbededVariableReference) prefix).getProgramVariable());
	} else if (prefix instanceof FieldReference) {
	    return convertToAttributeTerm
		((FieldReference) prefix);
	} else if (prefix instanceof ThisReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: this-references not supported yet");		
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else if (prefix instanceof SuperReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: super-references not supported yet");

	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else if (prefix instanceof MetaClassReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: metaclass-references not supported yet");
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else 	if (prefix instanceof ProgramVariable) {
	    // the base case: the leftmost item is a local variable
	    return termFab.createVariableTerm
		((ProgramVariable)prefix);
	} else 	if (prefix instanceof VariableReference) {
	    Debug.out("typeconverter: "+
		      "variablereference:", (((VariableReference)prefix).getProgramVariable()));
	    return termFab.createVariableTerm
		(((VariableReference)prefix).getProgramVariable());
	}  else if (prefix instanceof TypeReference) {	   
	    Function typeConst 
		= ((TypeReference)prefix).getKeYJavaType().typeConst();
	    return termFab.createFunctionTerm(typeConst);
	} else {
	    System.out.println("typeconverter: "+
		      "WARNING: unknown reference prefix:"+ prefix+ 
		      prefix.getClass());
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	}	
    }

    private Term convertToAttributeTerm(VariableReference fr) {	       
	ReferencePrefix prefix = fr.getReferencePrefix();
	Debug.out("TypeConverter: FieldReference: ",fr);
	Term sub = null;
	if (prefix==null && fr instanceof EmbededVariableReference) {
	    return termFab.createVariableTerm
		(((EmbededVariableReference)fr).getProgramVariable());
	}
	if (prefix != null && !(prefix instanceof PackageReference) ) {
	    sub = convertReferencePrefix(prefix);
	}
	if (sub != null) {
	    return termFab.createAttributeTerm(fr.getProgramVariable(),
					       sub.sort(), sub);
	} else {
	    System.out.println("typeconverter: Not supported reference type (fr, class):"+
		      fr+ fr.getClass());
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	    //return errorTerm;
	}
    }
    
    private Term convertNewOperator(New create) {
	OBJECT_COUNTER = OBJECT_COUNTER + 1;
        String typeName = create.getTypeReference().getName();
	Term t = termFab.createFunctionTerm
	    (new ProgramObject
		(new Name("obj_"+OBJECT_COUNTER+"|:"+typeName),
		 create.getTypeReference().getKeYJavaType()));
	return t;
    }

    public Term convertToLogicElement(ProgramElement pe) {
	return convertToLogicElement(pe, integerLDT, booleanLDT, objectLDT);
    }

    public Term convertToLogicElement(ProgramElement pe, 
				      LDT intLDT, LDT booleanLDT, 
				      LDT objectLDT) {
	Debug.out("typeconverter: called for:", pe, pe.getClass());
	if (pe instanceof ProgramVariable) {	    
	    return termFab.createFunctionTerm((ProgramVariable)pe);
	} else if (pe instanceof NullLiteral) {
	    return services.getJavaInfo().getNullConst();
	} else if (pe instanceof EmbededVariableReference) {
	    return convertToAttributeTerm((VariableReference)pe);
	//     return termFab.createFunctionTerm
// 		(((EmbededVariableReference)pe).getProgramVariable());
	} else if (pe instanceof FieldReference) {
	    return convertToAttributeTerm((FieldReference)pe);
	} else if (pe instanceof VariableReference) {
	    return termFab.createFunctionTerm
		(((VariableReference)pe).getProgramVariable());	      
	} else if (pe instanceof IntLiteral) {
	    return intLDT.translateLiteral((IntLiteral)pe);
	} else if (pe instanceof ParenthesizedExpression) {
	    return convertToLogicElement
		(((ParenthesizedExpression)pe).getChildAt(0), 
		 intLDT, booleanLDT, objectLDT); 
	} else if (pe instanceof New) {
	    return convertNewOperator((New) pe);
	} else if (pe instanceof LongLiteral) {
	    return intLDT.translateLiteral((LongLiteral)pe);
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof IntLiteral) {
	    String val = ((IntLiteral)((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new IntLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new IntLiteral("-"+val));
	    }
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof LongLiteral ) {
	    String val = ((LongLiteral)((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new LongLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new LongLiteral("-"+val));
	    }
	} else if (pe instanceof de.uka.ilkd.key.java.expression.Operator) {
	    return translateOperator
		((de.uka.ilkd.key.java.expression.Operator)pe, 
		 intLDT, booleanLDT, objectLDT);
	} else if (pe instanceof BooleanLiteral) {
	    return booleanLDT.translateLiteral((BooleanLiteral) pe);
	} else if (pe instanceof PrimitiveType) {
	    //	    return convertNewOperator((PrimitiveType) pe);
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this primitive type");
	    //return errorTerm;
	}	
	throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this logic element.");
	//return errorTerm;
    }

    private boolean isArithmeticOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof Divide || op instanceof Times || 
	    op instanceof Plus || op instanceof Minus || 
	    op instanceof Modulo || op instanceof ShiftLeft ||
	    op instanceof ShiftRight || op instanceof BinaryAnd || 
	    op instanceof BinaryNot || op instanceof BinaryOr || 
	    op instanceof BinaryXOr || op instanceof Negative || 
	    op instanceof PreIncrement || op instanceof PostIncrement ||
	    op instanceof PreDecrement || op instanceof PostDecrement) {
	    return true;
	} 
	return false;
    }


    private boolean isCompareOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof GreaterThan || op instanceof GreaterOrEquals ||
	    op instanceof LessThan || op instanceof LessOrEquals || 
	    op instanceof Equals || op instanceof LogicalNot
	    || op instanceof LogicalAnd || op instanceof LogicalOr) {
	    return true;
	}
	return false;
    }


    private boolean isBooleanOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof LogicalAnd || op instanceof LogicalOr ||
	    op instanceof LogicalNot) {
	    return true;
	}
	return false;
    }

    // this method performs binary numeric promotion on the arguments
    public KeYJavaType getPromotedType(KeYJavaType type1, KeYJavaType
				       type2) {
	if ((type1.getName().equals("boolean") &&
	     type1.getName().equals("boolean")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	if ((type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")) &&
	    (type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	if ((type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")  ||
	     type1.getName().equals("long")) &&
	    (type1.getName().equals("long")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	if ((type1.getName().equals("long")) &&
	    (type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")  ||
	     type1.getName().equals("int")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);	    
	throw new RuntimeException("Could not determine promoted type "+
				   "of "+type1+" and "+type2);
    }


    private Type resolveType
	(de.uka.ilkd.key.java.expression.Operator op) {	
	if (isArithmeticOperator(op)) {
	    return resolveArithmeticType(op);
	} else if (isCompareOperator(op)) {
	    return booleanLDT.javaType().head();
	} else if (isBooleanOperator(op)) {
	    return booleanLDT.javaType().head();
	}
	Debug.out("typeconverter: the return type of the given "+
		  "operator cannot be determined: ", op);
	return null;
    }

    private Type resolveArithmeticType
	(de.uka.ilkd.key.java.expression.Operator op) {	
	if (op instanceof BinaryAnd || 
	    op instanceof BinaryOr) {
	    return (getPrimitiveKeYJavaType((Expression)op.
					    getChildAt(0)).getJavaType());
	}
	// works to simple needs to be adjust in order 
	// to support real, double
	return integerLDT.javaType().head();
    }

    public Sort getPrimitiveSort(Type t) {
        LDT result = getModelFor(t);
        return (result == null ? null : result.targetSort());
    }


    public KeYJavaType getPrimitiveKeYJavaType(Expression e) {
	if (e instanceof ParenthesizedExpression) {
	    return getPrimitiveKeYJavaType
		((Expression)((ParenthesizedExpression)e).getChildAt(0));
	} else if (e instanceof New) {
	    return ((New)e).getTypeReference().getKeYJavaType();
	} else if (e instanceof NullLiteral) {
	    return services.getJavaInfo().getNullType();
	} else if (e instanceof MethodReference) {
	    KeYJavaType kjt = ((MethodReference)e).method(services.getJavaInfo())
		.getMethodDeclaration().getTypeReference().getKeYJavaType();
	    return kjt;	
	} else if (e instanceof Plus || e instanceof Minus || 
		   e instanceof Times || e instanceof Divide ||
		   e instanceof Modulo || e instanceof BinaryAnd ||
		   e instanceof BinaryOr || e instanceof BinaryXOr ||
		   e instanceof ShiftLeft || e instanceof ShiftRight ||
		   e instanceof UnsignedShiftRight) {	    
	    return getPromotedType
		(getPrimitiveKeYJavaType
		 ((Expression)
		  ((de.uka.ilkd.key.java.expression.Operator)e).getChildAt(0)),
		 getPrimitiveKeYJavaType
		 ((Expression)			
		  ((de.uka.ilkd.key.java.expression.Operator)e).getChildAt(1)));
	} else if (e instanceof IntLiteral) {
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	} else if (e instanceof LongLiteral) {
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	}  else if (e instanceof BooleanLiteral) {
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	}  else if (e instanceof Conditional) {
	    return getPrimitiveKeYJavaType(((Conditional)e).getExpressionAt(1)); 
                                       //%%%the 2nd expr should be involved, too!
	} else if (e instanceof ProgramVariable) {
	    return ((ProgramVariable) e).getKeYJavaType();
	} else if (e instanceof EmbededVariableReference) {
	    return ((EmbededVariableReference)e).getProgramVariable()
		.getKeYJavaType();
	} else if (e instanceof FieldReference) {
	    return ((ProgramVariable)((AttributeOp)convertToAttributeTerm
				((FieldReference)e).op()).attribute())
		                 .getKeYJavaType();
	} else if (e instanceof TypeCast) {
	    return ((TypeRef)((TypeCast)e).getChildAt(0)).getKeYJavaType();
	} else if (e instanceof PrimitiveType) {	    
	    return services.getJavaInfo().getKeYJavaType((PrimitiveType)e);
	} else if (e instanceof de.uka.ilkd.key.java.expression.Operator) {
	    return services.getJavaInfo().getKeYJavaType
		(resolveType((de.uka.ilkd.key.java.expression.Operator)e));
	} else {
	    Debug.out("Tried to get type for (expression, class)", e, e.getClass());
	}
	throw new RuntimeException("Cannot determine type of "+e.getClass()
				   +": "+e);
    }

    public Expression convertToProgramElement(Term term) {
	if (term.op() instanceof Function) {
	    for (int i=0; i<models.length; i++) {
		if (models[i].hasLiteralFunction((Function)term.op())) {
		    return models[i].translateTerm(term, null);	       
		}
	    }
	}
	ExtList children=new ExtList();
	for (int i=0; i<term.arity(); i++) {
	    children.add(convertToProgramElement(term.sub(i)));
	}
	if (term.op() instanceof ProgramInLogic) {
	    return ((ProgramInLogic)term.op()).convertToProgram(term, children);
	} else if (term.op() instanceof Function) {
	    for (int i=0; i<models.length; i++) {
		if (models[i].containsFunction((Function)term.op())) {
		    return models[i].translateTerm(term, children);
		}
	    }
	} 
	throw new RuntimeException("Cannot convert term to program: "+term
				   +" "+term.op().getClass());
    }

    public KeYJavaType getKeYJavaType(Term t) {
	if (t.op() instanceof ProgramInLogic) {
	    return ((ProgramInLogic)t.op()).getKeYJavaType();
	}
	return getPrimitiveKeYJavaType(convertToProgramElement(t));
    }

    public KeYJavaType getKeYJavaType(Type t) {    
	LDT model = getModelFor(t);
	if (model != null) {
	    return model.getKeYJavaType(t);
	} else {
	    Debug.out("javainfo: no predefined model for type ",t);
	    return null;
	}	
    }

}
