// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.ReferenceSuffix;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.gui.SequentPrinter;
import java.io.IOException;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramInLogic;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Visitor;
import de.uka.ilkd.key.logic.Term;

public class ProgramVariable extends TermSymbol 
                             implements SourceElement, ProgramElement, 
                             Expression, ReferencePrefix, IProgramVariable, 
                             ParsableVariable, ReferenceSuffix, ProgramInLogic {

    private final KeYJavaType type;

    public ProgramVariable(ProgramElementName name, KeYJavaType t) {
	this.name = name;
	this.type = t;
	this.sort = t.getSort(); 
    }

    public ProgramVariable(ProgramElementName name, Sort s) {
	this.name = name;
	this.type = null;
	this.sort = s; 
    }

    /** returns sort */
    public Sort sort() {
	return sort == null ? type.getSort() : sort;
    }

   /** @return arity of the Variable as int */
    public int arity() {
	return 0;
    }

    /** @return name of the TermSymbol */
    public Name name() {
	return name;
    }

    /** @return name of the ProgramVariable */
    public ProgramElementName getProgramElementName() {
	return (ProgramElementName) name;
    }
    
    /** toString */
    public String toString() {
	return name()+((sort == Sort.FORMULA)? "" : ":"+sort);
    }

    public SourceElement getFirstElement(){
	return this;
    }

    public SourceElement getLastElement(){
	return this;
    }

    public Comment[] getComments() {
	return null;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(de.uka.ilkd.key.java.visitor.Visitor v) {
	v.performActionOnProgramVariable(this);
    }


    /** the recoder pretty printer */
    public void prettyPrint(PrettyPrinter w) throws IOException {
	w.printProgramVariable(this);
    }

    /**
 *        Returns the start position of the primary token of this element.
 *        To get the start position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the start position of the primary token.
    */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the end position of the primary token of this element.
 *        To get the end position of the syntactical first token,
 *        call the corresponding method of <CODE>getLastElement()</CODE>.
 *        @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the relative position (number of blank heading lines and 
 *        columns) of the primary token of this element.
 *        To get the relative position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return  Position.UNDEFINED;
    }

    public KeYJavaType getKeYJavaType() {
	return type;
    }
    
    
    /**
     * We do not have a prefix, so fake it!
     * This way we implement ReferencePrefix
     * @author VK
     */
    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public void setReferencePrefix(ReferencePrefix r) {
    }
    
    
    /** equals modulo renaming is described in the corresponding
     * comment in class SourceElement. In this case two
     * programvariables are considered to be equal if they are
     * assigned to the same abstract name or if they are the same
     * object.
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	if (!(se instanceof ProgramVariable)) {
	    return false;
	}
	if (nat.contains(this)) {
	    return nat.sameAbstractName(this, se);
	}
	return (se == this);
    }

    public Expression convertToProgram(Term t, ExtList l) {
	return this;
    }

    
}
