// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.gui.SequentPrinter;
import java.io.IOException;
import de.uka.ilkd.key.logic.ProgramConstruct;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 * This class represents the root of a schema variable hierarchy to be
 * express termstructures that match on logical terms. Schema variables
 * are used in Taclets where they act as placeholders for other
 * TermSymbols. The TermSymbols a SchemaVariable is allowed to match
 * is specified by their type and sort. If a match fits these
 * conditions can be checked using the method isCompatible(TermSymbol t)
 */
public abstract class SchemaVariable extends TermSymbol
    implements ParsableVariable, QuantifiableVariable, ProgramConstruct {

    /** SchemaVariables are used as placeholder for other
     * TermSymbols. The allowed type of the symbols is stored here.
     */
    private Class matchType = TermSymbol.class;

    /** the instantiation of this variable in a given rule application
     *  is not 100% predetermined and worth saving in proof */
    private boolean interesting = false;
   

    /** creates a SchemaVariable representing quantified variables 
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the variable the SchemaVariable will be
     * used to represent
     * @return the SchemaVariable
     */
    public static SchemaVariable createVariableSV(Name name,
						  Sort sort) {
	return new VariableSV(name, sort);
    }

    /** this method tests on object identity
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	return se == this;
    }

    /** creates a SchemaVariable representing a term but not a formula
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the term the SchemaVariable will be
     * used to represent
     * @return the SchemaVariable
     */
    public static SchemaVariable createTermSV(Name name,
					      Sort sort) {
	return new TermSV(name, sort);
    }

    /** creates a SchemaVariable representing a formula 
     * @param name the Name of the SchemaVariable
     * @return the SchemaVariable
     */
    public static SchemaVariable createFormulaSV(Name name) {
	return new FormulaSV(name);
    }

    /** creates a SchemaVariable representing a program construct
     * @param name the Name of the SchemaVariable
     * @return the SchemaVariable
     */
    public static SchemaVariable createProgramSV(Name name, ProgramSVSort s){
	return new ProgramSV(name, s);
    }


    /** creates a new SchemaVariable. That is used as placeholder for
     * TermSymbols of a certain type.
     * @param name the Name of the SchemaVariable
     * @param matchType Class representing the type of symbols that
     * can be matched
     * @param sort the Sort of the SchemaVariable and the matched type     
     */    
    private SchemaVariable(Name name,
			  Class matchType,
			  Sort sort) {
	this.name = name;
	this.matchType = matchType;
	this.sort = sort;	
    }

//     never used, though possibly useful
//     /** checks if the given TermSymbol fits to the set of possible
//      * allowed matches
//      * @param symbol TermSymbol to be checked if the matching conditions
//      * (type and sort compatibility) are fulfilled
//      * @return true iff symbol fits the conditions
//      */
//     public boolean isCompatible(TermSymbol symbol) {
// 	return (symbol.sort() == sort() &&
// 		matchType.isInstance(symbol));
//     }

    public Class matchType() {
	return matchType;
    }

    /** returns true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     * @return true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     */
    public boolean isVariableSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a term but not a formula
     * @return true iff this SchemaVariable is used to match
     * a term but not a formula
     */
    public  boolean isTermSV() {
	return false;
    }


    /** returns true iff this SchemaVariable is used to match
     * a formula 
     * @return true iff this SchemaVariable is used to match
     * a formula
     */
    public boolean isFormulaSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a part of a program 
     * @return true iff this SchemaVariable is used to match
     * a part of a program
     */
    public boolean isProgramSV() {
	return false;
    }


   /** @return arity of the Variable as int */
    public int arity() {
	return 0;
    }

    /** toString */
    public String toString(String sortSpec) {
	return name+" ("+sortSpec+")"; 
    }

    
    // Stuff needed for ProgramSV 
    /** @returns comments if the schemavariable stands for programm
     * construct and has comments attached to it (not supported yet)
     */
    public Comment[] getComments() {
	return null;
    }

    public KeYJavaType getKeYJavaType() {
	return null;
    }

    public SourceElement getFirstElement(){
	return null;
    }

    public SourceElement getLastElement(){
	return null;
    }

    /**
 *        Returns the start position of the primary token of this element.
 *        To get the start position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the start position of the primary token.
    */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the end position of the primary token of this element.
 *        To get the end position of the syntactical first token,
 *        call the corresponding method of <CODE>getLastElement()</CODE>.
 *        @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the relative position (number of blank heading lines and 
 *        columns) of the primary token of this element.
 *        To get the relative position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return Position.UNDEFINED;
    }

    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public void setReferencePrefix(ReferencePrefix r) {
    }

    public int getDimensions(){
	return 0;
    }

    public int getTypeReferenceCount(){
	return 0;
    }

    public TypeReference getTypeReferenceAt(int index) {
	return this;
    }

    public PackageReference getPackageReference() {
	return null;
    }

    public int getExpressionCount() {
	return 0;
    }

    public Expression getExpressionAt(int index) {
	return null;
    }

    public int getChildCount() {
	return 0;
    }

    public ProgramElement getChildAt(int index) {
	return this;
    }
    
    public int getIndexOfChild(ProgramElement child) {
	return 0;
    }
    
    public int getChildPositionCode(ProgramElement child) {
	return 0;
    }

    public int getIndexOfChild(int positionCode) {
	return 0;
    }

    public int getRoleOfChild(int positionCode) {
	return 0;
    }

    public int getStatementCount() {
	return 0;
    }

    public int size() {
	return 0;
    }

    public ArrayOfExpression getUpdates() {
        return null;
    }

    public ArrayOfLoopInitializer getInits() {
	return null;
    }


    public Statement getStatementAt(int i) {
	return this;
    }

    public ProgramElementName getProgramElementName() {
	return new ProgramElementName(toString());
    }    

    public String getName() {
	return name.toString();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnSchemaVariable(this);
    }

    /** this pretty printer method is for the program pretty printer
     * and needs not to be overwritten by ProgramSV but at the moment
     * it is not 
     */
    public void prettyPrint(PrettyPrinter w) throws IOException {	
	w.printSchemaVariable(this);
    }


    public void markInteresting() {
       interesting = true;
    }

    public boolean isInteresting() {
       return interesting;
    }



////////////////////////////////////////////////////////////////////////

    /** inner class that represents only schemavariables that are
     * bound and therefore only match other variables and not terms
     */
    static private class VariableSV extends SchemaVariable {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * bound(quantified) variables.
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 */    
	private VariableSV(Name name,
			   Sort sort) {
	    super(name, QuantifiableVariable.class, sort); 	
	}

	/** returns true iff this SchemaVariable is used to match
	 * bound (quantifiable) variables 
	 * @return true iff this SchemaVariable is used to match
	 * bound (quantifiable) variables 
	 */
	public boolean isVariableSV() {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString("variable");
	}
    }

    /** inner class that represents schemavariables that are
     * used to match logical terms but not a formula.
     */
    static private class TermSV extends SchemaVariable {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * terms.
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 */    
	private TermSV(Name name,
		       Sort sort) {
	    super(name, Term.class, sort);	
	    if (sort == Sort.FORMULA) {
		throw new RuntimeException("A TermSV is not allowed to"
					   +" have the sort "+sort);
	    }

	}
	

	/** returns true iff this SchemaVariable is used to match
	 * a term but not a formula
	 * @return true iff this SchemaVariable is used to match
	 * a term but not a formula
	 */
	public boolean isTermSV() {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString(sort().toString()+" term");
	}
	
    }

    /** inner class that represents schemavariables that are
     * used to match a formula.
     */
    static private class FormulaSV extends SchemaVariable {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * formulae.
	 * @param name the Name of the SchemaVariable
	 */    
	private FormulaSV(Name name) {
	    super(name, Term.class, Sort.FORMULA);	
	}

	/** returns true iff this SchemaVariable is used to match
	 * a formula 
	 * @return true iff this SchemaVariable is used to match
	 * a formula
	 */
	public boolean isFormulaSV() {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString("formula");
	}	
    }

    /** inner class that represents schemavariables that are
     * used to match a part of a program
     */
    static class ProgramSV extends SchemaVariable {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * formulae.
	 * @param name the Name of the SchemaVariable
	 */    
	private ProgramSV(Name name, ProgramSVSort s) {
	    super(name, ProgramConstruct.class, s);	
	}

	/** returns true iff this SchemaVariable is used to match
	 * a formula 
	 * @return true iff this SchemaVariable is used to match
	 * a formula
	 */
	public boolean isProgramSV() {
	    return true;
	}

	/** @returns comments if the schemavariable stands for programm
	 * construct and has comments attached to it (not supported yet)
	 */
	public Comment[] getComments() {
	    return null;
	}

	public SourceElement getFirstElement(){
	    return this;
	}

	public SourceElement getLastElement(){
	    return this;
	}

	/**
	 * Returns the start position of the primary token of this element.
	 * To get the start position of the syntactical first token,
	 * call the corresponding method of <CODE>getFirstElement()</CODE>.
	 * @return the start position of the primary token.
	 */
	public Position getStartPosition(){
	    return Position.UNDEFINED;
	}

	/**
	 * Returns the end position of the primary token of this element.
	 * To get the end position of the syntactical first token,
	 * call the corresponding method of <CODE>getLastElement()</CODE>.
	 * @return the end position of the primary token.
	 */
	public Position getEndPosition(){
	    return Position.UNDEFINED;
	}

	/**
	 * Returns the relative position (number of blank heading lines and 
	 * columns) of the primary token of this element.
	 * To get the relative position of the syntactical first token,
	 * call the corresponding method of <CODE>getFirstElement()</CODE>.
	 * @return the relative position of the primary token.
	 */
	public Position getRelativePosition(){
	    return  Position.UNDEFINED;
	}

	/** toString */
	public String toString() {
	    return toString("program "+sort().name());
	}	
    }
}
