// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.parser.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;    
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.rule.*;

    import de.uka.ilkd.key.java.SchemaJavaReader;
    import java.util.Vector;
    import java.util.HashMap;
    import java.util.HashSet;
    import java.util.Iterator;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.proof.UseInsteadMap;
    import de.uka.ilkd.key.java.Services;

}

/**
 * A parser for Taclets, based on the term/formula grammar terms.g
 */  

class TacletParser extends TermParser;
options {
	importVocab=KeYLexer;
	k = 2;
	defaultErrorHandler=false;
}


{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected Namespace var_ns;
    protected Namespace func_ns;
    protected Namespace sort_ns;
    protected Namespace heur_ns;
    protected JavaInfo javaInfo;    

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;
    protected UseInsteadMap useInstead=new UseInsteadMap();

    protected TermFactory tf;
    protected SchemaJavaReader javaReader;
    boolean schematicJava;	

    public boolean errors = false;

    public TacletParser(TokenStream lexer,
		      String filename,
		      TermFactory tf,
		      SchemaJavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer,2);
	setFilename(filename);
	this.tf = tf;
	this.javaReader = jr;
	this.var_ns  = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;
	this.heur_ns = heur_ns;
	schematicJava=true;	
    }

    public TacletParser(TokenStream lexer,
		      String filename,
		      SchemaJavaReader jr,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename, TermFactory.DEFAULT, 
	     jr, var_ns, func_ns, sort_ns, heur_ns);
    }


    public TacletParser(TokenStream lexer,
		      String filename,
		      TermFactory tf,
		      JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename, tf, 
	     new de.uka.ilkd.key.java.SchemaRecoder2KeY(javaInfo.rec2key(),
						javaInfo.getTypeConverter()), 	     
	     var_ns, func_ns, sort_ns,heur_ns);
    }

    public TacletParser(TokenStream lexer,
		      String filename,
	              JavaInfo javaInfo,
		      Namespace var_ns,
		      Namespace func_ns,
		      Namespace sort_ns,
		      Namespace heur_ns) {
	this(lexer, filename,TermFactory.DEFAULT,
	     new de.uka.ilkd.key.java.SchemaRecoder2KeY(javaInfo.rec2key(),
						javaInfo.getTypeConverter()),
	     var_ns, func_ns, sort_ns, heur_ns);
    }


    private TacletBuilder createTacletBuilderFor(Object find) 
	throws InvalidFindException {
	if ( find == null ) {
	    return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
	    return new RewriteTacletBuilder().setFind((Term)find);
        } else if ( find instanceof Sequent ) {
	    Sequent findSeq = (Sequent) find;
	    if ( findSeq == Sequent.EMPTY_SEQUENT ) {
		return new NoFindTacletBuilder();
	    } else if (   findSeq.antecedent().size() == 1
			  && findSeq.succedent().size() == 0 ) {
		Term findFma = findSeq.antecedent().get(0).formula();
		return new AntecTacletBuilder().setFind(findFma);
	    } else if (   findSeq.antecedent().size() == 0
			  && findSeq.succedent().size() == 1 ) {
		Term findFma = findSeq.succedent().get(0).formula();
		return new SuccTacletBuilder().setFind(findFma);
	    } else {
		throw new InvalidFindException
		    ("Unknown find-sequent (perhaps null?):"+findSeq,
		     getFilename(), getLine(), getColumn());
	    }
        } else {
	    throw new RuntimeException("createTacletBuilderFor called with"
				       + find.getClass().getName() + "object");
        }
    }	    

	protected static HashSet progVars(JavaBlock jb) {
		ProgramVariableCollector pvc
			= new ProgramVariableCollector(jb.program());
		pvc.start();
		return pvc.result();
	}


	protected void bindProgVars(HashSet progVars) {
		var_ns=new Namespace(var_ns);
		Iterator it=progVars.iterator();
		while (it.hasNext()) {
			var_ns.add((ProgramVariable)it.next());
		}
	}

	protected void unbindProgVars() {
		var_ns = var_ns.parent();
	}



    private static ListOfRulevariable NO_RULE_VARIABLES =
	SLListOfRulevariable.EMPTY_LIST;

    private void addGoalTemplate(TacletBuilder b,
                                 String id,
				 Object rwObj,
				 Sequent addSeq,
				 ListOfTaclet addRList,
				 SetOfSchemaVariable pvs) 
	throws SemanticException
	{
	    TacletGoalTemplate gt = null;
	    if ( rwObj == null ) {
		// there is no replacewith, so we take
		gt = new TacletGoalTemplate(addSeq,
					  addRList,
					  NO_RULE_VARIABLES,
					  pvs);
	    } else {
		if ( b instanceof NoFindTacletBuilder ) {
		    // there is a replacewith without a find.
		    throw 
			new UnfittingReplacewithException
			("Replacewith without find", getFilename(),
			 getLine(), getColumn());
		} else if ( b instanceof SuccTacletBuilder
			    || b instanceof AntecTacletBuilder ) {
		    if ( rwObj instanceof Sequent ) {
			gt = new AntecSuccTacletGoalTemplate(addSeq,
							   addRList,
							   NO_RULE_VARIABLES,
							   (Sequent)rwObj,
							   pvs);  
		    } else {
			throw new UnfittingReplacewithException
			    ("Replacewith in a Antec-or SuccTaclet has "+
			     "to contain a sequent (not a term)", 
			     getFilename(), getLine(), getColumn());
		    }
		} else if ( b instanceof RewriteTacletBuilder ) {
		    if ( rwObj instanceof Term ) {
			gt = new RewriteTacletGoalTemplate(addSeq,
							 addRList,
							 NO_RULE_VARIABLES,
							 (Term)rwObj,
							 pvs);  
		    } else {
			throw new UnfittingReplacewithException
			    ("Replacewith in a RewriteTaclet has "+
			     "to contain a term (not a sequent)", 
			     getFilename(), getLine(), getColumn());
		    }
		}
	    }
            gt.setName(id); 
	    b.addTacletGoalTemplate(gt);
	}
						


    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine();
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int col = -1;
	try {
	    col = LT(0).getColumn();
	} 
	catch (Exception e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{
		if ( v instanceof LogicVariable ) {
			return tf.createVariableTerm((LogicVariable)v);
		} else if ( v instanceof SchemaVariable ) {
			return tf.createVariableTerm((SchemaVariable)v);
		} else if ( v instanceof ProgramVariable ) {
			return tf.createVariableTerm((ProgramVariable)v);
//		} else if ( v instanceof ParsableVariable ) {
//			return tf.createVariableTerm((ParsableVariable)v);

		} else {
			throw new antlr.SemanticException
			    (""+ v +" is not a logic or schema or program variable");
		}
	}

    protected TermSymbol progVar(Token id,Sort s) {
	TermSymbol v = (TermSymbol)
	    var_ns.lookup(new Name(id.getText()));
	if (v == null) {
	    v = new ProgramVariable(new ProgramElementName(id.getText()), s);
	    var_ns=var_ns.extended(v);
	}
	return v;
    }

    protected LogicVariable bindVar(Token id,Sort s) {
	LogicVariable v=new LogicVariable(new Name(id.getText()), s);
	var_ns=var_ns.extended(v);
	return v;
    }
	
    protected void unbindVars() {
	var_ns = var_ns.parent();
    }

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
	String s=t.getText();
	if (!schematicJava) {
		try {
			return javaReader.readBlockWithProgramVariables(var_ns, s);
		} catch (Exception e) {
			if (s.length()>10) {
				s=s.substring(0,10)+"...";
			}
			throw new antlr.SemanticException("Reading Java block "
				+ s + " failed. "
				+e.getMessage());	
		}
	}
	javaReader.setSVNamespace(var_ns);
	return javaReader.readBlockWithEmptyContext(s);
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	if (javaInfo==null) {
           javaInfo=javaReader.createJavaInfo(new Services());
        }
	return javaInfo;
    }
}

top {Taclet r;} : r=taclet;

taclet returns [Taclet r] 
{ 
	Sequent ifSeq = Sequent.EMPTY_SEQUENT;
	Object  find = null;
	r = null;
	TacletBuilder b = null;
}
	: 
		name:IDENT
		LBRACE
		( "if" LPAREN ifSeq=seq RPAREN ) ?
		( "find" LPAREN find=termorseq RPAREN ) ?
		{ 
			b = createTacletBuilderFor(find);
			b.setName(new Name(name.getText()));
			b.setIfseq(ifSeq);
		}
		( "varcond" LPAREN varexplist[b] RPAREN ) ?
		goalspecs[b]
		modifiers[b]
		RBRACE
		{ r = b.getTaclet(); }
    ;

modifiers[TacletBuilder b] :
    ( heuristics[b] 
    | "noninteractive" { b.setNoninteractive(true); }
    | "recursive"      { b.setRecursive(true); }
    | DISPLAYNAME dname:STRING_LITERAL 
                       {b.setDisplayName(dname.getText());}
    | HELPTEXT htext:STRING_LITERAL
                       {b.setHelpText(htext.getText());}
    ) *
    ;

seq returns [Sequent s] {Semisequent ant,suc; s = null; } : 
    ant=semisequent SEQARROW suc=semisequent 
    { s = Sequent.createSequent(ant, suc); }
    ;

termorseq returns [Object o]
{
	Term head = null;
	Sequent s = null;
	Semisequent ss = null;
	o = null; 
}
	:
		head=term ( COMMA s=seq | SEQARROW ss=semisequent ) ?
		{
			if ( s == null ) {
				if ( ss == null ) {
					// Just a term
					o = head;
				} else {
					// A sequent with only head in the antecedent.
					Semisequent ant = Semisequent.EMPTY_SEMISEQUENT;
					ant = ant.insertFirst(
						new ConstrainedFormula(head, Constraint.BOTTOM));
					o = Sequent.createSequent(ant,ss);
				}
			} else {
				// A sequent.  Prepend head to the antecedent.
				Semisequent newAnt = s.antecedent();
				newAnt = newAnt .insertFirst(
					new ConstrainedFormula(head, Constraint.BOTTOM));
				o = Sequent.createSequent(newAnt,s.succedent());
			}
		}
    |
		SEQARROW ss=semisequent
		{
			o = Sequent.createSequent(Semisequent.EMPTY_SEMISEQUENT,ss);
		}
    ;

semisequent returns [Semisequent ss]
{ 
	Term head = null;
	ss = Semisequent.EMPTY_SEMISEQUENT; 
}
    :
		/* empty */ | 
		head=term ( COMMA ss=semisequent ) ? 
		{ ss = ss.insertFirst(new ConstrainedFormula(head, Constraint.BOTTOM)); }
    ;

varexplist[TacletBuilder b] : varexp[b] ( COMMA varexp[b] ) * ;

varexp[TacletBuilder b]
{
	ParsableVariable x;
	ParsableVariable y;
	Sort s = null;
}
    :

		(( "typeof" LPAREN y=var_id  RPAREN  x=var_id "new") 
			{ b.addVarsNew((SchemaVariable) x, (SchemaVariable) y); }
		| (s=sort_id_check[true]  x=var_id "new") 
			{b.addVarsNew((SchemaVariable) x, s); }
		| (x=var_id "not" "free" "in" y=var_id)
			{ b.addVarsNotFreeIn((SchemaVariable)x,(SchemaVariable)y); }
		)
    ;

goalspecs[TacletBuilder b] :
		"close" "goal"
    | goalspec[b] ( SEMI goalspec[b] )* ;


goalspec[TacletBuilder b] 
{
	Object rwObj = null;
	Sequent addSeq = Sequent.EMPTY_SEQUENT;
	ListOfTaclet addRList = SLListOfTaclet.EMPTY_LIST;
	SetOfSchemaVariable addpv = SetAsListOfSchemaVariable.EMPTY_SET;
        String name = null;
}
    :
		(   ( "replacewith" (id:STRING_LITERAL)?
                      LPAREN rwObj=termorseq RPAREN
                         (addSeq=add)? 
                         (addRList=addrules)? 
			 (addpv=addprogvar)?
                    )
		| ( addSeq=add (addRList=addrules)? )
		| ( addRList=addrules )
		)
		{
                 	if (id != null) name = id.getText();
			addGoalTemplate(b,name,rwObj,addSeq,addRList,addpv);
		}
    ;

// $$$ has been inlined above to access the optional name
//replacewith returns [Object o] { o = null; } :
//		"replacewith" LPAREN o=termorseq RPAREN;

add returns [Sequent s] { s = null;} :
		"add" LPAREN s=seq RPAREN;

addrules returns [ListOfTaclet lor] { lor = null; } :
		"addrules" LPAREN lor=tacletlist RPAREN;

addprogvar returns [SetOfSchemaVariable pvs] {pvs = null; } :
                "addprogvars" LPAREN pvs=pvset RPAREN;

tacletlist returns [ListOfTaclet lor]
{ 
	Taclet head = null;
	lor = SLListOfTaclet.EMPTY_LIST; 
}
    :
		head=taclet   
		( /*empty*/ | COMMA lor=tacletlist ) { lor = lor.prepend(head); }
    ;

pvset returns [SetOfSchemaVariable pvs] 
{
        ParsableVariable pv = null;
	pvs = SetAsListOfSchemaVariable.EMPTY_SET;
}
    :
                pv=var_id
		( /*empty*/ | COMMA pvs=pvset ) { pvs = pvs.add
							((SchemaVariable)pv); };



heuristics[TacletBuilder b] :
		"heuristics" LPAREN heuristic[b] ( COMMA heuristic[b] ) * RPAREN ;

heuristic[TacletBuilder b] :
		id:IDENT
		{   
			Heuristic h = (Heuristic) heur_ns.lookup(new Name(id.getText()));
			if (h == null) {
				throw new NotDeclException("heuristic", id,
							   getFilename());
			}
			b.addHeuristic(h);
		}
	;


// This is overridden from terms.g to allow SchemaVariables in quantifiers.


bindingterm returns [Term a = null]
	:
		a = sv_or_local_bindingterm
	;

sv_or_local_bindingterm returns [Term a = null]
{
	Operator op = null;
	QuantifiableVariable v = null;
	TermSymbol ts, pv = null;
	Term a1,a2;
	Term a0=null;
	Sort s = null;
	boolean bound = false;
}
	:
		(( 
			"eps" { op = Op.EPSILON; }
		|	"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		v0:IDENT
		(   
			COLON s=sort_id 
			{
				v=bindVar(v0,s);
				bound = true;
			}
		|
			/*empty*/
			{
				ts = (TermSymbol) var_ns.lookup(new Name(v0.getText()));
				if (   ts == null
				    || ts instanceof LogicVariable ) {
					throw new SemanticException(
						"Quantified variables need a sort.");
				}
				if ( ! (ts instanceof SchemaVariable && 
					((SchemaVariable)ts).isVariableSV())) {
					throw new SemanticException(
						""+ts+" is not allowed in a quantifier.");
				}
				v = (SchemaVariable) ts;
				bound = false;
			}
		)
		DOT a1=term60 
		{ 
			if ( op == Op.EPSILON ) {
				a = tf.createEpsilonTerm(v,a1);
			} else {
				a = tf.createQuantifierTerm((Quantifier)op, v,a1);
			}
			if (bound) {unbindVars();}
		})
	|
        (
		LBRACE
		( 
		  (IDENT DOT IDENT ASSIGN)=>
		  (a0=attrterm ASSIGN a1=term60 RBRACE a2=term60
		      { 
			a = tf.createUpdateTerm(a0,a1,a2);			
		      }
		  ) |
		  (v1:IDENT 
			 ((ASSIGN term)=>(ASSIGN a1=term RBRACE
			      {	pv = progVar(v1,a1.sort());} a2=term60
			      { 
				if (pv instanceof ProgramVariable) {
					a = tf.createUpdateTerm((ProgramVariable)pv,a1,a2);
				} else {
					a = tf.createUpdateTerm((SchemaVariable)pv,a1,a2);
				} 
			      })|

  			  (COLON s=sort_id 
				{
				    bound = true;
				}
			  |
				/*empty*/
				{func_ns.lookup(new Name(v1.getText()))==null}?				
				{
				ts = (TermSymbol) var_ns.lookup(new Name(v1.getText()));
				if (   ts == null
				    || ts instanceof LogicVariable ) {
					throw new SemanticException(
						"Substituted variables need a sort.");
				}
				if ( ! (ts instanceof SchemaVariable &&
					((SchemaVariable)ts).isVariableSV()) ) {
					throw new SemanticException(
					"	"+ts+" is not allowed as bound variable "
						+"in a substitution term.");
				}
				v = (SchemaVariable) ts;
				bound = false;
			}
			a1=term RBRACE 
			{ 
			  if (bound) {v = bindVar(v1,s);} } a2=term60
			{ 	
			  a = tf.createSubstitutionTerm(Op.SUBST,v,a1,a2);
			  if (bound) {unbindVars();}
			}
			 
		   	)
			)
      		       ) 
		 )
		)
		
	;
		
// This is overridden from terms.g to allow SchemaVariables in attributes.



attrterm returns [Term a = null]
{	Vector vec=new Vector();
	ParsableVariable v=null;
	Term t=null;
	ParsableVariable v0=null;
    Sort so=null;

} : (t=funcpredvarterm {	so=t.sort();
        } (DOT^ v=attrid[so] {vec.add(v); so=v.sort(); })+) {
	if (!schematicJava) {
		for (int i=0; i<vec.size(); i++) {
			t=tf.createAttributeTerm((ProgramVariable)vec.get(i),
							t.sort(),t);
		}
	} else {
		t=tf.createAttributeTerm((SchemaVariable)v,t);
	}
	a=t;
    };


attrid [Sort s] returns [ParsableVariable v = null]
{

} : attr:IDENT 
{ if (!schematicJava) {
		v = getJavaInfo().getAttribute(attr.getText(), (ObjectSort)s);
	} else {
		v = (SchemaVariable)var_ns.lookup(new Name(attr.getText()));
	}

};



condterm returns [Term a = null]
{ Term a1=null;
  Term a2=null;
  Term a3=null;
  ParsableVariable v=null;
}
       : 
    LPAREN a1=term80 GUARD a2=term80 RPAREN DOT v=var_id CONDUP a3=term60 {
	AttributeOp aop=null;
	if (((SchemaVariable)v).sort()==ProgramSVSort.VARIABLE) {
	       aop=new AttributeOp((SchemaVariable)v);
	}	
	a = tf.createCondTerm(aop,a1,a2,a3);

}
;

