// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import javax.swing.table.AbstractTableModel;
import javax.swing.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import java.io.*;
import java.util.Vector;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.SetOfTacletApp;
import de.uka.ilkd.key.rule.IteratorOfTacletApp;
import de.uka.ilkd.key.rule.inst.*;

public class ApplyTacletDialogModel {
	
    /** the model used for the instantiation of the if-sequent */
    private IfChoiceModel[] ifChoiceModel;
    /** the model used for the instantiation of the schemavariables */
    private TacletInstantiationsTableModel tableModel;

    /** the application of the Taclet of which this model is used to
     * complete the instantiations of the schemavariables and/or
     * if-sequent
     */
    private TacletApp app;
    /** the sequent the application above is applied */
    private Sequent seq;

    /** listeners of this model */
    private Vector listeners = new Vector();
    /** the change event that is sent */
    private final ModelEvent changeEvent = new ModelEvent(this);

    /** namespace of variables */
    private Namespace var_ns;
    /** namespace of functions */
    private Namespace func_ns;
    /** namespace of sorts */
    private Namespace sort_ns;
    /** the java information object */
    private JavaInfo javaInfo;

    /**
     * create new data model for the apply Taclet dialog wrapping a combo box
     * model and a table model
     */
    public ApplyTacletDialogModel(TacletApp app, Sequent seq, JavaInfo javaInfo,
				Namespace var_ns, Namespace func_ns, 
				Namespace sort_ns) {
	this.app = app;
	this.seq = seq;
	this.javaInfo = javaInfo;
	this.var_ns = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;

	initIfChoiceModels();
	tableModel = new TacletInstantiationsTableModel
	    (app, javaInfo, var_ns, func_ns, sort_ns);
    }

    public void addModelChangeListener(ModelChangeListener l) {
	listeners.add(l);
    }

    public void removeModelChangeListener(ModelChangeListener l) {
	listeners.remove(l);
    }
	
    public IfChoiceModel ifChoiceModel(int i) {
	return ifChoiceModel[i];
    }

    public int ifChoiceModelCount() {
	return ifChoiceModel.length;
    }

    public TacletInstantiationsTableModel tableModel() {
	return tableModel;
    }

    public TacletApp application() {
	return app;
    }

    public Taclet taclet() {
	return (app.taclet());
    }

    public Term ifFma(int i) {
	return ifChoiceModel(i).ifFma();
    }

    private void initIfChoiceModels() {
 	Sequent ifseq = taclet().ifseq();
	int size = ifseq.antecedent().size()+ifseq.succedent().size();
	ifChoiceModel = new IfChoiceModel[size];
 	IteratorOfConstrainedFormula it = ifseq.iterator();
	for (int i=0; i<size; i++) {
	    ifChoiceModel[i] = 
		new IfChoiceModel(it.next().formula(), app, seq, 
				  javaInfo, var_ns, func_ns, sort_ns);
	}
    }

    /** creates the data model to be displayed in the dataTable */
    private TacletInstantiationsTableModel buildDataModel
	(SetOfTacletApp set) {
	de.uka.ilkd.key.util.Debug.assert(set.size()<=1, 
					  "Too many instantiations");
	IteratorOfTacletApp it = set.iterator();
	return new TacletInstantiationsTableModel(it.next(), javaInfo,
						var_ns, func_ns, sort_ns);
    }
    
    private ListOfSVInstantiations getInstMap(TacletApp app, Sequent seq){	
	return ((app.taclet()).matchIf(seq, app.instantiations(), javaInfo));
    }


    private Semisequent createUserBuildSemisequent(TacletApp app, int start, 
						   int size) 
	throws SVInstantiationException { 

	Semisequent semi = Semisequent.EMPTY_SEMISEQUENT;
	for (int i = start; i < start+size; i++) {  	    	    
	    Term t = ifChoiceModel(i).getSelection();
       	    if (t!=null) {
		semi = semi.insertLast(new ConstrainedFormula(t)); 
	    } else {
		throw new SVInstantiationException("Nothing entered as manual"
						   +" input.","",0,0,null);
	    }
	}
	return semi;
    }

    private Sequent createUserBuildSequent(TacletApp app) 	
	throws SVInstantiationException{

	int endAntec = (app.taclet()).ifseq().antecedent().size();
	return Sequent.createSequent
	    (createUserBuildSemisequent(app, 0, endAntec),
	     createUserBuildSemisequent (app, endAntec,  
	      (app.taclet()).ifseq().succedent().size()));  
    }

    private TacletApp createTacletAppFromIfs(TacletApp app) 
	throws SVInstantiationException {

	Sequent userBuildIfSequent = createUserBuildSequent(app);
	ListOfSVInstantiations listOfSchemaTermMappings = 
	    getInstMap(app, userBuildIfSequent);
	de.uka.ilkd.key.util.Debug.assert(listOfSchemaTermMappings.size() <= 1,
					  "ambiguous instantiations");
 	if (listOfSchemaTermMappings.head() == null) {
 	    throw new IfMismatchException("If-formulae "
 					  +"mismatch.");
 	}	
	return app.addInstantiation(listOfSchemaTermMappings.head());

    }

    public String getStatusString() {
	TacletApp rapp=app;
	try {
	    rapp=createTacletApp();
	} catch(Exception ie) {
	    return "Rule cannot be applied. "+ie.getMessage();
	}
	if (rapp==null) {
	    return "Rule cannot be applied";
	}
	return "Instantiation is OK.";
	
    }

    public TacletApp createTacletApp() throws SVInstantiationException {       
	return createTacletAppFromIfs(tableModel.createTacletAppFromVarInsts());
    }
    
    
    private void informListenerAboutModelChange() {
	for (int i = 0; i < listeners.size(); i++) {
	    if (listeners.get(i) != null) {
		((ModelChangeListener)listeners.get(i)).modelChanged(changeEvent);
	    }
	}
    }

    /** sets the manual if-input */
    public void setManualInput(int i, String s) {
	ifChoiceModel(i).setInput(s);
	informListenerAboutModelChange();
    }

    /**
     * replaces the TacletApp of this ApplyTacletDialogModel by an TacletApp
     * where all name conflicts are resolved and thus the parser is enabled
     * to accept variables from the context or the prefix of the Taclet.
     *
     */
    public void prepareUnmatchedInstantiation() {
	app = app.prepareUserInstantiation();	
    }

    public Namespace var_ns() {
	return var_ns;
    }

}
