// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* represents a proof goal */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.IteratorOfProgramVariable;
import de.uka.ilkd.key.logic.op.SetOfProgramVariable;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

public class Goal implements java.io.Serializable {

    private Node node;

    /** the index of possible rule applications at this node
     * @undirected
     * @supplierRole possible rule applications
     * @clientCardinality 1
     * @supplierCardinality 1*/
    private RuleAppIndex ruleAppIndex;

    /** list of all applied ruleapps at this branch */
    private ListOfRuleApp appliedRuleApps = SLListOfRuleApp.EMPTY_LIST;

    /** creates a new goal referencing the given node */
    public Goal(Node node, RuleAppIndex ruleAppIndex, 
		ListOfRuleApp appliedRuleApps) {  
	this.node = node;
	this.ruleAppIndex = ruleAppIndex;
	this.appliedRuleApps = appliedRuleApps;
	//	node.addNoPosTacletApps(ruleAppIndex.tacletIndex().allNoPosTacletApps());
    }

    /** creates a new goal referencing the given node */
    public Goal(Node node, RuleAppIndex ruleAppIndex) {  
	this(node, ruleAppIndex, SLListOfRuleApp.EMPTY_LIST);
    }

    /** returns the referenced node */
    public Node node() {
	return node;
    }

    public SetOfProgramVariable getGlobalProgVars() {
	return node().getGlobalProgVars();
    }

    /**
     * adds the global program variables to a new created variable namespace
     * that contains all the elements of the given namespace.
     */
    public Namespace getVariableNamespace(Namespace exNS) {
	Namespace newNS=exNS;
	IteratorOfProgramVariable it=getGlobalProgVars().iterator();
	if (it.hasNext()) {
	    newNS=newNS.extended(it.next());
	}
	while (it.hasNext()) {
	    newNS.add(it.next());
	}
	return newNS;
    }

    public void setGlobalProgVars(SetOfProgramVariable s) {
	node.setGlobalProgVars(s);
    }

    /** set the node the goal is related to
     * @param node the new Node the goal belongs to
     */
    public void setNode(Node node) {
	this.node=node;
    }

    public void setRuleAppIndex(RuleAppIndex ruleAppIndex) {
	this.ruleAppIndex=ruleAppIndex;
    }

    /** returns the index of possible rule applications at this node
     */
    public RuleAppIndex ruleAppIndex() {
	return ruleAppIndex;
    }
     
    /**
     * returns the Taclet index for this goal. This is just a shortcut to the
     * Taclet index of the RuleAppIndex
     * @return the Taclet index assigned to this goal
     */
    public TacletIndex indexOfTaclets() {
	return ruleAppIndex.tacletIndex();
    }
   
    /** adds a formula to the sequent before the given position
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be added
     * @param p PosInOccurrence encodes the position 
     */
    public void addFormula(ConstrainedFormula cf, PosInOccurrence p) {
	setSequent(seq().addFormula(cf, p));
	ruleAppIndex().formulaAdded(cf, 
				    new PosInOccurrence(cf,
							PosInTerm.TOP_LEVEL, 
							seq()));   
    }

    /** adds a formula to the antecedent or succedent of a
     * sequent. Either at its front or back.
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be added
     * @param inAntec boolean true(false) if ConstrainedFormula has to be
     * added to antecedent (succedent) 
     * @param first boolean true if at the front, if false then cf is
     * added at the back
     */
    public void addFormula(ConstrainedFormula cf, boolean inAntec,
			  boolean first) {
	Sequent seq=seq();
	Semisequent semi=(inAntec ? seq.antecedent() : seq.succedent());
	semi=(first ? semi.insertFirst(cf) : semi.insertLast(cf));
	seq=Sequent.createSequent((inAntec ? semi : seq.antecedent()), 
				  (inAntec ? seq.succedent() : semi));
	setSequent(seq);

	ruleAppIndex().formulaAdded(cf, 
				    new PosInOccurrence(cf,
							PosInTerm.TOP_LEVEL, 
							seq));   
    }

    /** returns set of rules applied at this branch 
     * @return ListOfRuleApp applied rule applications
     */
    public ListOfRuleApp appliedRuleApps() {
	return appliedRuleApps;
    }
	

    /** returns the sequent of the node 
     * @return the Sequent to be proved
     */
    public Sequent seq() {
	return node().sequent();
    }

    /** sets the sequent of the node 
     * @param the Sequent to be set
     */
    public void setSequent(Sequent seq) {
	node().setSequent(seq);
    }

    /** adds a formula to the sequent after the given position
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be added
     * @param p PosInOccurrence encodes the position 
     */
    public void addFormulaAfter(ConstrainedFormula cf, PosInOccurrence p) {
	setSequent(seq().addFormulaAfter(cf, p));
	ruleAppIndex().formulaAdded(cf, 
				    new PosInOccurrence(cf,
							PosInTerm.TOP_LEVEL, 
							seq()));
    }

    /** replaces a formula at the given position  
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula replacing the old one
     * @param p PosInOccurrence encodes the position 
     */
    public void changeFormula(ConstrainedFormula cf, PosInOccurrence p) {
	setSequent(seq().changeFormula(cf, p));
	ruleAppIndex().formulaChanged(cf, p);
    }

    /** removes a formula at the given position from the sequent 
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be removed
     * @param p PosInOccurrence encodes the position 
     */
    public void removeFormula(PosInOccurrence p) {
	setSequent(seq().removeFormula(p));
	ruleAppIndex().formulaRemoved(p);
    }

    /**
     * puts the NoPosTacletApp to the set of TacletApps at the node
     * of the goal and to the current RuleAppIndex.
     * @param app the TacletApp
     */
    public void addNoPosTacletApp(NoPosTacletApp app) {
	node().addNoPosTacletApp(app);
	ruleAppIndex.addNoPosTacletApp(app);	
    }

    /**
     * creates a new TacletApp and puts it to the set of TacletApps at the node
     * of the goal and to the current RuleAppIndex.
     * @param rule the Taclet of the TacletApp to create
     * @param insts the given instantiations of the TacletApp to be created
     */
    public void addTaclet(Taclet rule, SVInstantiations insts) {		
	NoPosTacletApp tacletApp = NoPosTacletApp.createNoPosTacletApp(rule, insts);
	if (tacletApp != null) {
	    addNoPosTacletApp(tacletApp);
	}
    }

    public void addProgramVariable(ProgramVariable pv) {
	node.setGlobalProgVars(getGlobalProgVars().add(pv));
    }

    public void addProgramVariables(Namespace ns) {
	IteratorOfNamed it=ns.elements().iterator();
	while (it.hasNext()) {
	    addProgramVariable((ProgramVariable)it.next());
	}
    }


    /** clones the goal (with copy of tacletindex and ruleAppIndex)
     * @return Object the clone
     */
    public Object clone() {	
	return new Goal(node, ruleAppIndex.copy(), appliedRuleApps);
    }

    /** like the clone method but returns right type
     * @return Goal clone of this Goal
     */
    public Goal copy() {
	return (Goal)clone();
    }

    /**
     * puts a RuleApp to the list of the applied rule apps at this goal
     * and stores it in the node of the goal
     * @param app the applied rule app
     */
    public void addAppliedRuleApp(RuleApp app) {
	appliedRuleApps = appliedRuleApps.append(app);
// 	if (appliedRuleApps.size()>10) {
// 	    appliedRuleApps = appliedRuleApps.tail();
// 	}
	node().setAppliedRuleApp(app);
    }

    
    /** creates n new nodes as children of the
     * referenced node and new 
     * n goals that have references to these new nodes.
     * @return the list of new created goals.
     */
    public ListOfGoal split(int n) {
	ListOfGoal goalList=SLListOfGoal.EMPTY_LIST;
        if (n>0) {
	    Node parent = node(); // has to be stored because the node
				// of this goal will be replaced
	    Node newNode = null;
	    Goal newGoal = null;

	    for (int i=0;i<n;i++) {
		if (i==0) { // first new goal is this one 
		    newGoal = this;
		} else { // otherwise it is a copy
		    newGoal = copy();
		}

		// create new node and add to tree
		newNode = new Node(parent.proof(), parent.sequent(),
				   null, parent);
		// newNode.addNoPosTacletApps(parent.getNoPosTacletApps());
		newNode.setGlobalProgVars(parent.getGlobalProgVars());
		parent.add(newNode);

		// make new Goal and add to list
		newGoal.setNode(newNode);
		goalList = goalList.prepend(newGoal);
	    }
	}
	return goalList;
    }

    /**
     * sets back the proof step that led to this goal. This goal is set to
     * the parent node of the node corresponding to this goal. Goals given in
     * the goal list parameter are removed from that list, if their
     * corresponding nodes are leaves of the parent node of this goal.
     * @param goallist the list of goals 
     * @return the new list of goals where goals mapped to the leaves of
     * the parent to this goal are removed from compared to the given list.
     */
    public ListOfGoal setBack(ListOfGoal goalList) {
	Node parent=node.parent();
	IteratorOfNode leavesIt=parent.leavesIterator();
	IteratorOfGoal goalIt;
	while (leavesIt.hasNext()) {
	    Node n=leavesIt.next();
	 
	    goalIt=goalList.iterator();
	    while (goalIt.hasNext()) {
		Goal g=goalIt.next();
	
		if (g.node()==n && g!=this) {
		    goalList=goalList.removeAll(g);
		   
		}
	    }
	}

	//	ruleAppIndex.tacletIndex().setTaclets(parent.getNoPosTacletApps());

	ruleAppIndex.tacletIndex().removeTaclets(node.getNoPosTacletApps());
	setGlobalProgVars(parent.getGlobalProgVars());

	IteratorOfNode siblings=parent.childrenIterator();
	Node[] sibls=new Node[parent.childrenCount()];
	int i=0;
	while (siblings.hasNext()) {
	    sibls[i]=siblings.next(); 
	    i++;
	}

	for (i=0; i<sibls.length; i++) {
	    sibls[i].remove();
	}
	parent.setAppliedRuleApp(null);
	setNode(parent);
	return goalList;
    }
    
    public void setName(String s) {
       node.setName(s);
    }


    /** toString */
    public String toString() {
	String result=(node.sequent().prettyprint().toString());
	return result;
    }
}
