// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.util.List;
import java.util.Iterator;
import java.util.LinkedList;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;


public class Node implements java.io.Serializable {
    /** the proof the node belongs to */
    private Proof proof;
    private Sequent seq=Sequent.EMPTY_SEQUENT;
    private List children=new LinkedList();
    private Node parent=null;
    private RuleApp appliedRuleApp;
    private SetOfProgramVariable globalProgVars
 	= SetAsListOfProgramVariable.EMPTY_SET;
    private String name = null;
        
    /**
     * @supplierCardinality 1 
     */
    private SetOfNoPosTacletApp setOfNoPosTacletApp
	=SetAsListOfNoPosTacletApp.EMPTY_SET;
    
    /** creates an empty node that is root and leaf.
     */
    public Node(Proof proof) {
	this.proof = proof;
    }    

    /** creates a node with the given contents
     */
    public Node(Proof proof, Sequent seq) {
	this.proof = proof;
	this.seq=seq;
    }


    /** creates a node with the given contents, the given collection
     * of children (all elements must be of class Node) and the given
     * parent node.
     */
    public Node(Proof proof, Sequent seq, List children, Node parent) {
	this.proof = proof;
	this.seq=seq;	
	this.parent=parent;
	if (children!=null) this.children=children;
    }

    /** sets the sequent at this node
     */
    public void setSequent(Sequent seq) {
	this.seq=seq;
   } 

    /** returns the sequent of this node */
    public Sequent sequent() {
	return seq;
    }

    /** returns the proof the Node belongs to */
    public Proof proof() {
	return proof;
    }    

    public void setAppliedRuleApp(RuleApp ruleApp) {
	this.appliedRuleApp=ruleApp;
    }


    public RuleApp getAppliedRuleApp() {
	return appliedRuleApp;
    }
    
    /** Returns the set of NoPosTacletApps at this node */
    public SetOfNoPosTacletApp getNoPosTacletApps() {
	return setOfNoPosTacletApp;
    }

    public SetOfProgramVariable getGlobalProgVars() {
	return globalProgVars;
    }

    public void setGlobalProgVars(SetOfProgramVariable progVars) {
	globalProgVars=progVars;
    }

     /**
      * adds a new NoPosTacletApp to the set of available NoPosTacletApps 
      * at this node
      */
     public void addNoPosTacletApp(NoPosTacletApp s) {
 	setOfNoPosTacletApp = setOfNoPosTacletApp.add(s);
     }

    /**
     * adds several new NoPosTacletApps to the set of available 
     * NoPosTacletApps at this node
     * @param tacletAppSet the NoPosTacletApps to be added
     */
//     public void addNoPosTacletApps(SetOfNoPosTacletApp tacletAppSet) {
// 	long t = System.currentTimeMillis();
// 	setOfNoPosTacletApp = setOfNoPosTacletApp.union(tacletAppSet);
// 	System.out.println("Time For set union:"+(t-System.currentTimeMillis()));

//     }


    /** returns the parent node of this node.
     */
    public Node parent() {
	return parent;
    }

    /** returns true, iff this node is a leaf, i.e. has no children.
     */
    public boolean leaf() {
	return children.size()==0;
    }

    /** searches for a given node in the subtree starting with this node */
    public boolean find(Node node) {
	if (node == this) {
	    return true;
	}
	for (int i = 0; i < children.size(); i++) {
	    if (((Node)children.get(i)).find(node)) {
		return true;
	    }
	}
	return false;
    }

    /**  returns true, iff this node is root, i.e. has no parents.
     */
    public boolean root() {
	return parent==null;
    }  

    /** makes the given node a child of this node.
     */
    public void add(Node child) {
	children.add(child);
	child.parent = this;
	proof.fireProofExpanded(this);
    }

    /** removes child/parent relationship between this node and its
     * parent; if this node is root nothing happens.
     */
    public void remove() {
	if (parent != null) {
	    parent.remove(this);
	} 
    }

    /** removes child/parent relationship between the given node and
     * this node; if the given node is not child of this node,
     * nothing happens and then and only then false is returned.
     * @return false iff the given node was not child of this node and
     * nothing has been done.
     */
    public boolean remove(Node child) {
	if (children.remove(child)) {
	    child.parent = null;
	    proof().fireProofPruned(this, child);
	    return true;
	} else {
	    return false;
	}
    } 

    /** computes and returns the leaves of the subtree below this
     * node. Only for use in leavesIterator.
     */
    private List leaves() {
	List leaves = new LinkedList();
	if (leaf()) {
	    leaves.add(this);
	} else {	
	    IteratorOfNode childrenIt=childrenIterator();
	    while (childrenIt.hasNext()) {
		leaves.addAll(childrenIt.next().leaves());
	    }
	}
	return leaves;
    }

    /** returns an iterator for the leaves of the subtree below this
     * node. The computation is called at every call!
     */
    public IteratorOfNode leavesIterator() {
	return new NodeIterator(leaves().iterator());
    }

    /** returns an iterator for the direct children of this node.
     */
    public IteratorOfNode childrenIterator() {
	return new NodeIterator(children.iterator());
    }

    /** returns number of children */
    public int childrenCount() {
	return children.size();
    }

    /** returns i-th child */
    public Node child(int i) {
	return (Node)(children.get(i));
    }


    /** helps toString method 
     * @param prefix needed to keep track if a line has to be printed
     * @param tree the tree representation we want to add this subtree
     " @param preEnumeration the enumeration of the parent without the
     * last number
     * @param postNr the last number of the parents enumeration 
     * @param maxNr the number of nodes at this level
     * @param ownNr the place of this node at this level
     */
 
    private StringBuffer toString(String prefix, 
				  StringBuffer tree, 
				  String preEnumeration,
				  int postNr,
				  int maxNr,
				  int ownNr
				  ) {       
	IteratorOfNode childrenIt = childrenIterator(); 	
	// Some constants
	String frontIndent=(maxNr>1 ? " " : "");
	String backFill="   "; // same length as connectNode without
			       // frontIndent 
	String connectNode=(maxNr>1 ? frontIndent+"+--" : "");	
	String verticalLine=(maxNr>1 ? frontIndent+"|"+backFill : " |");
	

	// get enumeration
	String newEnumeration=preEnumeration;
	int newPostNr=0;
	if (maxNr>1) {
	    newEnumeration+=postNr+"."+ownNr+".";	    
	    newPostNr=1;
	} else {
	    newPostNr=postNr+ownNr;
	}
		
	// node is printed
	
	if (postNr!=0) { // not starting node (usually not root)
	    // prefix is appended twice in order to get an
	    // empty line between two nodes
	    tree.append(prefix); 
	    tree.append(verticalLine);
	    tree.append("\n");
       	    tree.append(prefix);	    
	    // indent node
	    tree.append(connectNode);
	} 
       
	tree.append("("+newEnumeration+newPostNr+") "+sequent().toString()+"\n");	

	// create new prefix
	if (ownNr<maxNr) { 
	    // connect node with next node of same level
	    prefix+=verticalLine;
	} else if (ownNr==maxNr && maxNr>1) {
	    // last node of level no further connection
	    prefix+=frontIndent+" "+backFill;
	} else if (ownNr!=maxNr && maxNr<=1) {
	    prefix="";
	}

	// print subtrees
	int childId=0;
	while (childrenIt.hasNext()) {		    
	    childId++;
	    childrenIt.next().toString(prefix, tree, newEnumeration, 
				       newPostNr,
				       children.size(), childId);
	}

	return tree;
    }


    public String toString() {
	StringBuffer tree=new StringBuffer();
	return "\n"+toString("",tree,"",0,0,1); 
    }
    
    
    public void setName(String s) {
       name = s;
    }
    
    public String name() {
      return name;
    }

    

    /**
     * checks if the parent has this node as child and continues recursively
     * with the children of this node.
     * @return true iff the parent of this node has this node as child and
     * this condition holds also for the own children.
     */
    public boolean sanityCheckDoubleLinks() {
	if (!root()) {
	    if (!parent().children.contains(this)) {
		return false;
	    }
	    if (parent.proof() != proof()) {
		return false;
	    }
	}
	if (!leaf()) {
	    IteratorOfNode it=childrenIterator();
	    while (it.hasNext()) {
		if (!it.next().sanityCheckDoubleLinks())
		    return false;
	    }
	}
	return true;
    }

    // inner iterator class 

    private class NodeIterator implements IteratorOfNode {
	private Iterator it;
	
	NodeIterator(Iterator it) {
	    this.it=it;
	}

	public boolean hasNext() {
	    return it.hasNext();
	}

	public Node next() {
	    return (Node)it.next();
	}
    }
}
