// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.util.Iterator;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;
import java.util.HashSet;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Name;

public class Proof implements Named, java.io.Serializable {

    /** name of the proof */
    private Name name;

    /** the root of the proof */
    private Node root;

    /** list with prooftree listeners of this proof */
    private List listenerList = Collections.synchronizedList(new ArrayList(10));

    /** list with the open goals of the proof */ 
    private ListOfGoal openGoals = SLListOfGoal.EMPTY_LIST;


    /** constructs a new empty proof */
    public Proof() {
	this(new Name(""));
    }

    /** constructs a new empty proof with name */
    public Proof(Name name) {
	this.name = name;
    }

    /** constructs a new empty proof with name */
    public Proof(String name) {
	this(new Name(name));
    }

    public Name name() {
	return name;
    }
    
    public Node root() {
	return root;
    }

    /** sets the root of the proof */
    public void setRoot(Node root) {
	if (this.root != null) {
	    throw new IllegalStateException
		("Tried to reset the root of the proof.");
	} else {
	    this.root = root;
	    fireProofStructureChanged();
	}
    }

    /** Sets up an initial proof. The given node will become the root and 
     * first goal. The given TacletIndex contains all available rules
     * @param root the Node to be the root of the proof 
     * @param index the TacletIndex with all available Taclets
     */
    public void setUp(Node root, TacletIndex index) {
	setRoot(root);
	Goal firstGoal = 
	    new Goal(root(), 
		     new RuleAppIndex
		     (new TacletAppIndex(index))
		     );
	openGoals = openGoals.prepend(firstGoal);
    }
    

    /** returns the list of open goals
     * @return list with the open goals
     */
    public ListOfGoal openGoals() {
	return openGoals;
    }

    /** removes the given goal and adds the new goals in list 
     * @param goal the old goal that has to be removed from list
     * @param newGoal the ListOfGoal with the new goals that were
     * result of a rule application on goal
     */
    public void replace(Goal oldGoal, ListOfGoal newGoals) {
	openGoals = openGoals.removeAll(oldGoal);
	if (newGoals == SLListOfGoal.EMPTY_LIST && 
	    closed()) {
	    fireProofClosed();
 	} else {
	    fireProofGoalRemoved(oldGoal);
	    add(newGoals);	
	}
    }

    /** removes the given goal from the list of open goals. Take care
     * removing the last goal will fire the proofClosed event
     * @param goal the Goal to be removed
     */
    public void remove(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.removeAll(goal);
	if (newOpenGoals != openGoals) {
	    openGoals = newOpenGoals;	
	    if (closed()) {
		fireProofClosed();
	    } else {
		fireProofGoalRemoved(goal);
	    }
	}
    }

    /** adds a new goal to the list of goals 
     * @param goal the Goal to be added 
     */
    public void add(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.prepend(goal);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	    fireProofGoalsAdded(goal);
	}
    }

    /** adds a list with new goals to the list of open goals 
     * @param goals the ListOfGoal to be prepended 
     */
    public void add(ListOfGoal goals) {
	ListOfGoal newOpenGoals = openGoals.prepend(goals);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	}

	// For the moment it is necessary to fire the message ALWAYS
	// in order to detect branch closing.
	fireProofGoalsAdded(goals);
	
    }

    /** returns true if the goal list is empty and therefore the proof
     * is closed 
     * @return true if the goal list is empty and therefore the proof
     * is closed
     */
    public boolean closed() {
	return (openGoals == SLListOfGoal.EMPTY_LIST);   
    }

    /** returns true iff sets back to the step that created the given
     * goal. If the undo operation was succesful true is returned.
     * @param goal the Goal desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public  boolean setBack(Goal goal) {		
	if (goal != null) {
	    Node parent = goal.node().parent();
	    if (parent != null) {
		openGoals = goal.setBack(openGoals);
		fireProofGoalsChanged();
		return true;
	    } 
	}
	// root reached or proof closed
	return false;
    }

    /** sets back the subtree of node including the node 
     *	itself if the subtree is not closed yet.
     * @param node the node desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public boolean setBack(Node node) {
	if (node == root) {
	    return false;
	}
	Goal goal = getGoal(node);
	if (goal!=null) {
	    return setBack(goal);
	} else {
	    ListOfGoal goalList = getSubtreeGoals(node);
	    if (goalList != SLListOfGoal.EMPTY_LIST) {
		// get different parents		
		HashSet parentSet = new HashSet();
		ListOfGoal removeList = SLListOfGoal.EMPTY_LIST;
		IteratorOfGoal goalIt = goalList.iterator();
		while (goalIt.hasNext()) {
		    Goal nextGoal = goalIt.next();
		    if (!parentSet.contains(nextGoal.node().parent())) {
			removeList = removeList.prepend(nextGoal);
			parentSet.add(nextGoal.node().parent());
		    }
		}
		// iterate through removeList
		IteratorOfGoal removeIt = removeList.iterator();
		while (removeIt.hasNext()) {
		    setBack(removeIt.next());
		}
		return setBack(node);
	    }
	    return false;
	}
    }


    public void updateProof() {
	fireProofGoalsChanged();
    }


    /** fires the event that the proof has been expanded at the given node */
    protected synchronized void fireProofExpanded(Node node) {
	ProofTreeEvent e = new ProofTreeEvent(this, node);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofExpanded(e);
	    }
	}
    } 

    /** fires the event that the proof has been pruned at the given node */
    protected synchronized void fireProofPruned(Node node, Node removedNode) {
	ProofTreeEvent e = new ProofTreeRemovedNodeEvent(this, node, removedNode);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofPruned(e);
	    }
	}
    } 

    /** fires the event that the proof has been restructured */
    protected synchronized void fireProofStructureChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofStructureChanged(e);
	    }
	}
    } 

    /** fires the event that a goal has been removed from the list of goals */
    protected synchronized void fireProofGoalRemoved(Goal goal) {
	ProofTreeEvent e = new ProofTreeEvent(this, goal);
	synchronized(listenerList) {
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofGoalRemoved(e);
	    }	
	}
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected synchronized void fireProofGoalsAdded(ListOfGoal goals) {
	ProofTreeEvent e = new ProofTreeEvent(this, goals);
	synchronized(listenerList) {
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofGoalsAdded(e);
	    }	
	}
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected synchronized void fireProofGoalsAdded(Goal goal) {
	fireProofGoalsAdded(SLListOfGoal.EMPTY_LIST.prepend(goal));
    }

    /** fires the event that the proof has been restructured */
    protected synchronized void fireProofGoalsChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofStructureChanged(e);
	    }
	}
    } 

    /** fires the event that the proof has closed. 
     * This event fired instead of the proofGoalRemoved event when
     * the last goal in list is removed.
     */
    protected synchronized void fireProofClosed() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofClosed(e);
	    }	
	}
    }

    public synchronized void addProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.add(listener);
	}
    }

    public synchronized void removeProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.remove(listener);
	}
    }

    /** returns true if the given node is part of a Goal 
     * @return  true if the given node is part of a Goal 
     */
    public boolean isGoal(Node node) {	
	return getGoal(node) != null;
    }

    /** returns the goal that belongs to the given node or null if the
     * node is an inner one 
     * @return the goal that belongs to the given node or null if the
     * node is an inner one 
     */
    public Goal getGoal(Node node) {	
	Goal result = null;
	IteratorOfGoal it = openGoals.iterator();
	while (it.hasNext()) {
	    result = it.next();
	    if (result.node() == node) {
		return result;
	    }
	}
	return null;
    }

    /** returns the list of goals of the subtree starting with node 
     * @param node the Node where to start from
     * @return the list of goals of the subtree starting with node 
     */
    public ListOfGoal getSubtreeGoals(Node node) {	
	ListOfGoal result = SLListOfGoal.EMPTY_LIST;
	IteratorOfGoal goalsIt  = openGoals.iterator();
	while (goalsIt.hasNext()) {
	    Goal goal = goalsIt.next();
	    IteratorOfNode leavesIt = node.leavesIterator();
	    while (leavesIt.hasNext()) {
		if (leavesIt.next() == goal.node()) {
		    result = result.prepend(goal);
		}
	    }
	}
	return result;
    }

    /** returns true iff the given node is found in the proof tree 
     *	@param node the Node to search for
     *	@return true iff the given node is found in the proof tree 
    */
    public boolean find(Node node) {
	if (root == null) {
	    return false;
	}
	return root.find(node);
    }

    /** toString */
    public String toString() {
	StringBuffer result = new StringBuffer();
	    result.append("Proof -- ");
	    if (!"".equals(name.toString())) {
		result.append(name.toString());
	    } else {
		result.append("unnamed");
	    }
	result.append("\nProoftree:\n");
	result.append(root.toString());
	return result.toString();
    }
}
