// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** this class extends TacletApp that have
 * no position information.
 */
package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

public class NoPosTacletApp extends TacletApp {

    /** creates a NoPosTacletApp for the given taclet and no instantiation
     * information and CHECKS variable conditions as well as it resolves
     * collisions
     * @param taclet the Taclet 
     */
    public static NoPosTacletApp createNoPosTacletApp(Taclet taclet) {
	return new NoPosTacletApp(taclet);
    }

    /** creates a NoPosTacletApp for the given taclet with some known instantiations
     * and CHECKS variable conditions as well as it resolves
     * collisions
     * @param taclet the Taclet 
     * @param instantiations the SVInstantiations
     */
    public static NoPosTacletApp createNoPosTacletApp(Taclet taclet, 
					    SVInstantiations instantiations) {
	SVInstantiations inst = resolveCollisionVarSV(taclet, instantiations);
	if (checkVarCondNotFreeIn(taclet, inst)) {
	    return new NoPosTacletApp(taclet, inst);
	} 
	return null;
    }

    /** creates a NoPosTacletApp for the given taclet and no instantiation
     * information
     * @param taclet the Taclet 
     */
    private NoPosTacletApp(Taclet taclet) {
	super(taclet);
    }

    /** creates a NoPosTacletApp for the given taclet with some known instantiations
     * @param taclet the Taclet 
     * @param instantiations the SVInstantiations
     */
    private NoPosTacletApp(Taclet taclet, SVInstantiations instantiations) {
	super(taclet, instantiations);
    }


    /** checks if the variable conditions of type 'x not free in y' are
     * hold by the found instantiations. The variable conditions is used
     * implicit in the prefix. (Used to calculate the prefix)
     * @param taclet the Taclet that is tried to be instantiated. A match for the
     * find (or/and if) has been found.
     * @param instantiations the SVInstantiations so that the find(if) matches
     * @return true iff all variable conditions x not free in y are hold
     */
    protected static boolean checkVarCondNotFreeIn(Taclet taclet,
					 SVInstantiations instantiations) {
	IteratorOfSchemaVariable it = instantiations.svIterator();
	while (it.hasNext()) {
	    SchemaVariable sv = it.next();


	    if (!sv.isProgramSV()) {
		if (!sv.isVariableSV()) {		
		    SetOfQuantifiableVariable boundVarSet = 
			boundAtOccurrenceSet(taclet.getPrefix(sv), 
					     instantiations);
		    if (!taclet.getPrefix(sv).context() && 
			!((Term)instantiations.getInstantiation(sv)).freeVars().
			subset(boundVarSet)) {		    
			return false;
		    }
		}
	    }
	}
	return true;
    }


    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return the new TacletApp
     */
    public TacletApp addInstantiation(SchemaVariable sv, Term term) {
	return createNoPosTacletApp(taclet(), 
				  instantiations().add(sv, term));
    }


    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param pe the ProgramElement the SV is instantiated with
     * @return the new TacletApp
     */
    public TacletApp addInstantiation(SchemaVariable sv, ProgramElement pe) {
	return createNoPosTacletApp((FindTaclet)taclet(), 
				instantiations().add(sv, pe));
    }



    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations and hold the old ones
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    public TacletApp addInstantiation(SVInstantiations svi) {
	return new NoPosTacletApp(taclet(), svi.union(instantiations()));
    }


    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations and forget the ones in this TacletApp
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    public TacletApp setInstantiation(SVInstantiations svi) {
	return new NoPosTacletApp(taclet(), svi);
    }


    /** returns true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     * @return true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     */
    public boolean complete() {
	return (uninstantiatedVars() == SetAsListOfSchemaVariable.EMPTY_SET
		&& taclet() instanceof NoFindTaclet);

    }

    protected SetOfQuantifiableVariable contextVars(SchemaVariable sv) {
	return SetAsListOfQuantifiableVariable.EMPTY_SET;
    }


    /**
     * returns the PositionInOccurrence (representing a ConstrainedFormula and
     * a position in the corresponding formula)
     * @return the PosInOccurrence
     */
    public PosInOccurrence posInOccurrence() {
	return null;
    }

    /** checks if term and NoPosTacletApp fit in some particular conditions */
    public MatchResult checkInteractive(Term term, 
					ListOfHeuristic heuristics,
					JavaInfo javaInfo) {
	if (taclet() instanceof FindTaclet) {
	    return ((FindTaclet)taclet()).checkInteractive
		(term, heuristics, instantiations(), javaInfo);
	} else {
	    return ((NoFindTaclet)taclet()).checkInteractive
		(heuristics, instantiations(), javaInfo);
	}

    }


    /** checks if term and NoPosTacletApp fit in some particular conditions */
    public MatchResult checkHeuristic(Term term,
				      ListOfHeuristic heuristics,
				      JavaInfo javaInfo) {
	if (taclet() instanceof FindTaclet) {
	    return ((FindTaclet)taclet()).checkHeuristic(term, heuristics, 
						     instantiations(), 
						     javaInfo);
	} else {
	    return ((NoFindTaclet)taclet()).checkHeuristic(heuristics,
						       instantiations(), 
						       javaInfo);
	}
    }



}
