// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.util;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import de.uka.ilkd.key.util.make.Config;
import de.uka.ilkd.key.util.install.*;
import java.util.jar.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

/**
 * A small installer class, to install key on windows systems.
 * reads in some path information for key and libraries
 * 
 * This class substitutes several scripts written for the installation
 * procedure: install.sh, install.key, createConfig
 *
 * The class is compatible with both the windows and linux common installation 
 * procedure. It has not been checked wether it is compatible with the
 * installation procedure for the developer version of the KeY-Project.
 */

public class InstallKeY {

    public static String DEV = "dev";
    public static String COMMON = "common";

    private String program;
    private String welcome;
    private String programpath;
    private String programpackage; //always ends with '/' well File.Separator


    /* together j */
    private String togetherlocal;
    private String togetherhome;
    private String togetherversion;

    /**
     * A table for saving path information for libraries used by the project.
     */
    private Vector paths;
    
    private double percentDone=0.0;
    private String currentAction="";
    private ProgressAdapter progressListener=null;
    

    /**
     * The constructor for the installer class.
     * @param w a welcome message
     * @param p the name of the file containing the executables (a jar file!)
     * @param prefix a string array indicating the package structure for the
     *        project
     */
    public InstallKeY(String w,String p,String[] prefix){
	this.welcome = w;
	this.program = p;
	this.programpath = new String();
	this.paths = new Vector();
	this.programpackage = "";
	for(int i=0;i<prefix.length;i++){
	    this.programpackage = this.programpackage 
		+ prefix[i] + File.separator;
	}
    }
    
    
    /** sets the progress listener */
    public void addProgressListener(ProgressAdapter adapter) {
	this.progressListener=adapter;
    }

    /**
     * A small wrapper for a BufferedReader, which reads input
     * from the keyboard.
     * @param lib the file for which path info is expected
     * @param msg a message displayed to the user for further information
     * @return the string that was typed in by the user.
     */
    public String readPath(String lib, String msg, String last){
	System.out.println("Please enter the directory of "
			   +lib+" "+msg+":");

	String path = new String();
	try{
	    BufferedReader in = 
		new BufferedReader(new InputStreamReader(System.in));
	    path = in.readLine();
	}
	catch(IOException e){ 
	    System.out.println(e);
	}

	return "".equals(path) ? last : path;
    }

    /**
     * A method for adding more libraries to the path-table. First
     * calls readPath to obtain path information from the user, then
     * writes the obtained information to the path-table.
     * @param lib the library for which path info is needed.
     * @param path the path to the lib
     * @param dev true if lib is only necessary for developer version
     */
    public void addLib(String lib, String path, boolean dev){
	String[] libinfo = new String[3];
	libinfo[0] = (dev ? DEV : COMMON);
	libinfo[1] = lib;
	libinfo[2] = path;
	this.paths.add(libinfo);
    }

    /**
     * This method is used to enter the path for the mainprogram.
     * This method only calls the readPath method.
     * @param msg an explanatory message given to the user.
     */
    public void enterMainPath(String msg){
	this.programpath = 
	    readPath(this.program, msg, "")+File.separator+"key.jar";
    }

    public void enterTogetherPaths(){
	this.togetherhome = readPath("Togetherhome","", "");
	this.togetherlocal = readPath("Togetherlocal",
				      "directory for key configuration files", "");
	this.togetherversion = readPath("Togetherversion", "5.0,5.5 or 6.0", "");
    }

    /**
     * This method creates a file called config.mk. This file is
     * later used when the configuration in key.util.make.config is
     * called.
     * @param configfilename the name of the configfile to be created
     */
    public void createConfig(String configfilename){
	String common = this.getCommonPath();
	String dev = this.getDevPath();

	/* not sure if needed outside developer dist */
	String jikes = "not available";
	    //"\$(JAVA_HOME)/jre/lib/rt.jar:\$(JAVA_HOME)/jre/lib/i18n.jar";

	java.util.Properties sys = System.getProperties();
	String javahome = sys.getProperty("java.home");

	String config = "CLASSPATH_COMMON[\n"+common+"]\n"+
	    "CLASSPATH_DEV[\n"+dev+"]\n"+
	    "JIKESPATH[\n"+jikes+"]\n"+ 
	    "LOCALTOGETHER[\n"+this.togetherlocal+"]\n"+
	    "LOCALTOGETHERCONFIG[\n"+this.togetherlocal+
	    File.separator+"config"+File.separator+"]\n"+	    
	    "KEYPATTERNPATH[\n"+this.togetherlocal+
	       File.separator+
	       "keypatterns/de/uka/ilkd/key/casetool/together/patterns/".
	       replace('/',File.separatorChar)+"]"+
	    "KEYSCRIPTPATH[\n"+this.togetherlocal+File.separator+"keyscripts]"+
	    "TOGETHERHOME[\n"+this.togetherhome+"]\n"+
	    "KEYJAR[\n"+this.programpath+"]\n"+
	    "JDK[\n"+javahome+"]\n";
	
	try{
	    FileWriter fw = new FileWriter(configfilename);
	    fw.write(config);
	    fw.close();
	}
	catch(IOException e){System.out.println(e);}
    }

    /**
     * Creates the path for the libraries for the common 
     * installation only.
     * @return returns the string representing that path
     */
    public String getCommonPath(){ 
	String common = new String(this.programpath);
	for(int i=0;i<this.paths.size();i++){
	    String[] libinfo = (String[])this.paths.get(i);
	    if(libinfo[0].equals(COMMON)){
		common += ("".equals(common) ? "": File.pathSeparator)+ 
		    libinfo[2];
		if (new File(common).isDirectory()) {
		    common += File.separator+libinfo[1];
		}
	    }
	}
	return common;
    }


    /**
     * Creates a a string of the paths needed for the developer version
     * @return a string with paths
     */
    public String getDevPath(){ 
	String dev = new String("");
	for(int i=0;i<this.paths.size();i++){
	    String[] libinfo = (String[])this.paths.get(i);
	    if(libinfo[0].equals(DEV)){
		dev = dev + libinfo[2]+File.separator+libinfo[1]+File.pathSeparator;
	    }
	}
	return dev + 
	    this.togetherhome+"/lib/openapi.jar"+
	    File.pathSeparator+
	    this.togetherhome+"/lib/jgl.zip"+
	    File.pathSeparator+
	    this.togetherhome+"/lib/together.jar"+
	    File.pathSeparator+
	    this.togetherhome+"/modules";
    }

    public void fireProgressChanged() {
	if (progressListener!=null) {
	    progressListener.progressChanged(currentAction, 
					     percentDone); 
	}
    }

    public void setPercentDone(double value) {
	this.percentDone=value;
	fireProgressChanged();
    }

    public void setCurrentAction(String value) {
	this.currentAction=value;
	fireProgressChanged();
    }

    public String getCurrentAction() {
	return currentAction;
    }

    public double getPercentDone() {
	return percentDone;
    }


    /**
     * Creates several directories needed for configuration
     */
    public String createDirs(){
	
	String keyscripts=this.togetherlocal+File.separator+
	    "keyscripts"+File.separator;
	String keypatterns=this.togetherlocal+File.separator +
	    "keypatterns"+File.separator+this.programpackage+
	    "casetool"+File.separator+"together";

	String[] paths=new String[]{
	    togetherlocal, keyscripts, 
	    keyscripts+"menuextension",
	    keyscripts+"patternupdate", 
	    keypatterns, 
	    this.togetherlocal+File.separator+"bin",
	    this.togetherlocal+File.separator+"config"	    
	};


	boolean success = true;
	for (int i=0; i<paths.length; i++) {
	    File file=new File(paths[i]); 
	    if (!file.exists()) {
		success = success && file.mkdirs();
		if (!success) {
		    return "Error creating: "+file; 
		}
	    }
	    setPercentDone((double)(i/(paths.length-1)));
	}
	
	// no error occured
	return null;
    }

    
    /**
     * Extracts files from the key/program jar-archive.
     * @param entryname name of the file in the jar
     * @param writetopath  where to copy     
     */
    public void extractFromJar(String entrypath,
			       String writetopath){
	try {
	    //generate directories to write file:
	    boolean dirs = (new File(writetopath)).mkdirs();

	    String jarfilename = this.programpath;
	    // Open the JAR file
	    JarFile jarfile = 
		new JarFile(jarfilename);
	    Enumeration entries = jarfile.entries();

	    while (entries.hasMoreElements()) {

		JarEntry entry = (JarEntry) entries.nextElement();
		if (entry.getName().startsWith(entrypath)) {
		    if (entry.isDirectory()) {
			//createDir
			new File(writetopath+entry.getName()).mkdirs();
		    } else{		   
			InputStream in = jarfile.getInputStream(entry);
			//Write to file
			FileOutputStream out = new FileOutputStream(writetopath+entry.getName());
			int c;
			while ((c = in.read()) != -1) {
			    out.write(c);
			}
			out.close();
			in.close();
		    }
		}
	    }
	} catch (Exception e) {
	    System.err.println(e.toString());
	    e.printStackTrace();
	}
    }


    /**
     * Extracts files from the key/program jar-archive.
     * @param entryname name of the file in the jar
     * @param filename name to be copied to.
     */
    public void extractFromJar(String entrypath,
			       String writetopath,
			       String filename){
	try {
	    //generate directories to write file:
	    boolean dirs = (new File(writetopath)).mkdirs();

	    String jarfilename = this.programpath;
	    // Open the JAR file
	    JarFile jarfile = new JarFile(jarfilename);

	    // get JarEntry
	    System.out.println("Extracting "+filename+" from "+jarfilename);
	    JarEntry entry = (JarEntry)jarfile.getEntry(entrypath+filename);
	    if (entry == null) {
		System.err.println("Entry "+entrypath+filename+
				   "not found in jar file.");
		System.exit(-1);
	    }
	    InputStream in = jarfile.getInputStream(entry);
	    //Write to file
	    FileOutputStream out = new FileOutputStream(writetopath+filename);
	    int c;
	    while ((c = in.read()) != -1) {
		out.write(c);
	    }
	    out.close();
	    in.close();
	} catch (Exception e) {
	    System.err.println(e.toString());
	    e.printStackTrace();
	}
    }
    
    /**
     * Copies a directory to another directory. 
     * Subdirectories are copied as well
     * @param from the source directory
     * @param to the target directory
     */
    public void copyDir(String from, String to) 
	throws FileNotFoundException, IOException {

	String srcdir = from;
	String distdir = to;
	String[] filename = (new File(srcdir)).list();

	
	setPercentDone(0.0);
	
	for(int i=0; i<filename.length; i++){
	    FileReader in =new FileReader(srcdir+File.separator+filename[i]);
	    FileWriter out=new FileWriter(distdir+File.separator+filename[i]);
	    int c;
	    while ((c = in.read()) != -1) {
		out.write(c);
	    }
	    in.close();
	    out.close();
	    setPercentDone((double)i/(filename.length-1));
	}	
    }

    /**
     * Appends file2 to file 
     * @param file1 first file
     * @param file2 second file
     */
    public void concat(String file1, String file2){
	try{
	    File misc = new File(file1);
	    File ext = new File(file2);
	    File result = new File(file1+".tmp");

	    FileInputStream miscin = new FileInputStream(misc);
	    FileInputStream extin = new FileInputStream(ext);
	    SequenceInputStream in = new SequenceInputStream(miscin, extin);
	    FileOutputStream out = new FileOutputStream(result);
	    int c;
	    while ((c = in.read()) != -1) {
		out.write(c);
	    }
	    in.close();
	    out.close();
	    misc.delete();
	    result.renameTo(misc);
	}
	catch(FileNotFoundException e)
	    {System.err.println("File Not Found");}
	catch(IOException e)
	    {System.err.println("Concatenation failed.");}
    }

    /**
     * Copies directory with together config files (takes some time!)
     * Copies several of the generated files to other directories
     * And concatenates config files.
     */
    public void copyFiles() 
	throws FileNotFoundException, IOException {
	
	setPercentDone(0.0);
	setCurrentAction("Copying config files");
	this.copyDir(this.togetherhome+File.separator+"config",
		     this.togetherlocal+File.separator+"config");

	/* moves generated script 'startkey' to local/bin dir */
	setCurrentAction("Generating startscripts");
	boolean renamed = 
	    (new File("startkey")).renameTo(new File(this.togetherlocal
						     +File.separator+"bin"
						     +File.separator
						     +"startkey"));
	setPercentDone(0.5);
	renamed = 
	    (new File("startkey.bat")).renameTo(new File(this.togetherlocal
						     +File.separator+"bin"
						     +File.separator
						     +"startkey.bat"));
	setPercentDone(1.0);

	/* moves 'configDefaultSnapShot' to local dir */
	setPercentDone(0.0);
	setCurrentAction("Set up default configs");

	String name="configDefaultSnapShot";
	String newname=this.togetherlocal+File.separator+"bin"
	    +File.separator+"configDefaultSnapShot.config";
	renamed = (new File(name)).renameTo(new File(newname));

	setPercentDone(0.5);

	//overkill ? --> concatenate 'miscConfigExtension' and 'misc.config'
	this.concat(this.togetherlocal+File.separator+"config"
		    +File.separator+"misc.config","miscConfigExtension");
	setPercentDone(1.0);

	// Data for filenames:
	String prefix = this.togetherlocal+File.separator+"keyscripts"
	    +File.separator+"menuextension"+File.separator;
	String menu =this.programpackage+"casetool"
	    +File.separator+"together"+File.separator+"scripts"
	    +File.separator+"menuextension"+File.separator;
	String pattern=this.programpackage+"casetool"
	    +File.separator+"together"+File.separator+"scripts"
	    +File.separator+"patternupdate"+File.separator;
	String file="Manifest.mf";

	String mentry = 
	    "de/uka/ilkd/key/casetool/together/scripts/menuextension/";
	String pentry =
	    "de/uka/ilkd/key/casetool/together/scripts/patternupdate/";

	setPercentDone(0.0);
	setCurrentAction("Installing together plug-ins");
	this.extractFromJar(mentry,prefix+menu,file);  
	setPercentDone(0.5);
	this.extractFromJar(pentry,prefix+pattern,file);  		
	setPercentDone(1.0);

	// copy patterns
// 	String commonPatternPrefix=
// 	    "de/uka/ilkd/key/casetool/together/patterns/JAVA/";
// 	String dest=this.togetherlocal+File.separator+
// 	    "keypatterns"+File.separator;
// 	this.extractFromJar(commonPatternPrefix+"CLASS/", dest);
// 	this.extractFromJar(commonPatternPrefix+"MEMBER/", dest);
    }

    /**
     * This calls the key-config-Utility.
     * @param configfile the name of the created configfile
     * @param installtype the type of installation procedure indicates
     *        the files that have to be generated, f.e. 'dist' or 'dev'
     */
    public void callKeyConfig(String configfile, String installtype){
	// arguments for key configuration utility:
	String cargs[] = new String[3];
	cargs[0] = configfile;
	cargs[1] = installtype;
	cargs[2] = this.togetherversion;

	//call key configuration utility -> creates some files.
	de.uka.ilkd.key.util.make.Config.main(cargs);
    }

    private static void abort(String error) {
	System.err.println(error);
	System.exit(-1);
    }

    private static void guiAbort(String error) {
	JOptionPane.showMessageDialog(null, error, "Installation aborted", 
				      JOptionPane.ERROR_MESSAGE);
	System.exit(-1);
    }

    public static void guiInstall(InstallKeY win) {
	Installer installer=new Installer("KeY-Installer");

	JLabel header=new JLabel();
	header.setFont(new Font("Header Font", Font.PLAIN, 24));
	header.setText("KeY-Installer");
	header.setIcon(de.uka.ilkd.key.gui.IconFactory.keyLogo(50,40));
	header.setHorizontalAlignment(SwingConstants.CENTER);
	installer.setHeader(header);
	installer.addInstallationPane(new WelcomePane());
	installer.addInstallationPane(new GlobalSettingsPane());
	installer.addInstallationPane(new LibrarySettingsPane());
	installer.pack();
	installer.setVisible(true);
	

	while (installer.exitCode()==Installer.RUNNING) {
	    Thread.yield();
	}

	if (installer.exitCode()==Installer.CANCEL) {
	    guiAbort("Installation aborted by user.");
	} else {
	    System.out.println("Reading information...");

	    win.addProgressListener(new GUIProgressAdapter());
	    win.setCurrentAction("Creating Dirs");

	    for (int i=0; i<installer.paneCount(); i++) {
		System.out.println("Processing:"+installer.pane(i).name());
		if ("Global".equals(installer.pane(i).name())) {
		    GlobalSettingsPane gsp=
			(GlobalSettingsPane)installer.pane(i);
		    win.programpath = gsp.path(0);
		    win.togetherlocal = gsp.path(1);
		    win.togetherhome  = gsp.path(2);
		    win.togetherversion  = "6.0";
		} else if ("Library".equals(installer.pane(i).name())) {
		    LibrarySettingsPane lsp=
			(LibrarySettingsPane)installer.pane(i);
		    win.addLib("antlr.jar", lsp.path(0) ,false);
		    win.addLib("dresden-ocl-demo.jar", lsp.path(1), false);
		    win.addLib("xerces-1_4_4.jar", lsp.path(2),false);
		    win.addLib("recoder.jar", lsp.path(3), false);
		    win.addLib("xml.jar", lsp.path(4), false);
		}
	    }
	    //creating files for configuration of key/together
	    win.createConfig("config.mk");
	    win.callKeyConfig("config.mk","dist");

	    //formerly the task of the 'install.key' script
	    String errorStr;
	    if ((errorStr=win.createDirs())!=null) {
		guiAbort(errorStr);
	    } 
	    try {
		win.copyFiles();
	    } catch(FileNotFoundException fne) {		
		errorStr="Tried to copy non-existant file.";
		errorStr+="Received the following exception:\n"+fne;
		errorStr+="Aborted Installation due to error.";		
		guiAbort(errorStr);
		System.exit(-1);
	    } catch(IOException e) {
		errorStr="Error copying file. Possible reasons:\n"+
		    " * not enough memory \n * no permission";
		errorStr+="Received the following exception:\n"+e;
		errorStr+="Aborted Installation due to error.";		
		guiAbort(errorStr);
		System.exit(-1);
	    }
	    //finally finished.
	    

	    JOptionPane.showMessageDialog
		(null, "Installation Completed.\n"+
		 "Linux/Unix: Please execute: \n chmod u+x "+win.togetherlocal+
		 File.separator+"bin"+File.separator+"startkey");
	}
	System.exit(0);       
    }

    public static void cmdLineInstall(InstallKeY win) {
	//enter all the important paths:
	win.enterMainPath("(usually the directory where you unpacked KeY.tgz)");
	win.enterTogetherPaths();
	
	//third param is true if lib is needed for dev-version only
	String libpath = win.readPath("antlr.jar","", "");
	win.addLib("antlr.jar", libpath ,false);
	
	libpath = win.readPath("dresden-ocl-demo.jar", libpath, libpath);
	win.addLib("dresden-ocl-demo.jar", libpath, false);
	
	libpath = win.readPath("xerces-1_4_4.jar", libpath, libpath);
	win.addLib("xerces-1_4_4.jar", libpath ,false);
	
	libpath = win.readPath("recoder.jar", libpath, libpath);
	win.addLib("recoder.jar", libpath, false);
	
	libpath = win.readPath("xml.jar", libpath, libpath);
	win.addLib("xml.jar", libpath, false);

	//creating files for configuration of key/together
	win.createConfig("config.mk");
	win.callKeyConfig("config.mk","dist");

	//formerly the task of the 'install.key' script
	System.out.println("Creating directory hierarchy");
	String errorStr=win.createDirs();
	if ((errorStr=win.createDirs())!=null) {
	    abort(errorStr);
	} 

	System.out.println("Copying config files (this may take some time)");
	try {
	    win.copyFiles();
	} catch(FileNotFoundException fne) {		
	    errorStr="Tried to copy non-existant file.";
	    errorStr+="Received the following exception:\n"+fne;
	    errorStr+="Aborted Installation due to error.";		
	    abort(errorStr);
	    System.exit(-1);
	} catch(IOException e) {
	    errorStr="Error copying file. Possible reasons:\n"+
		" * not enough memory \n * no permission";
	    errorStr+="Received the following exception:\n"+e;
	    errorStr+="Aborted Installation due to error.";		
	    abort(errorStr);
	    System.exit(-1);
	}
	//finally finished.
	System.out.println("Unix/Linux-User: Please execute:\n"+
			   "\t chmod u+x "+win.togetherlocal+File.separator+
			   "bin"+File.separator+"startkey");
	System.out.println("Installed.");
	System.exit(0);
    }

    /**
     * The main program. This is an example implementation of an installation
     * procedure for the key-project. First some have to be entered by the 
     * user. From that information some configuration files are created either
     * by this class or by the key-config-utility. Finally the local together
     * directory structure is created and the necessary files are copied there.
     * This used to be done by the generated script 'install.key'. 
     * @param args so far no arguments are needed.
     * @see de.uka.ilkd.key.util.make.Config
     */
    public static void main(String[] args){
	String welcome = "Welcome!\nDirectory names require ABSOLUTE PATHS.\n";
	//
	String[] prefix = {"de","uka","ilkd","key"};
	InstallKeY win = new InstallKeY(welcome,"key.jar",prefix);
	boolean gui = true;
	if (args.length>0 && "--text".equals(args[0])) {
	    gui=false;
	} else {
	    try {
		if (GraphicsEnvironment.
		    getLocalGraphicsEnvironment().getScreenDevices() == null ||
		    GraphicsEnvironment.
		    getLocalGraphicsEnvironment().getScreenDevices().length==0) {
		    gui = false;
		}
	    } catch(Throwable err) {
		gui = false;
	    }
	}

	if (gui) {
	    guiInstall(win);
	} else {
	    cmdLineInstall(win);
	}
// 	libpath = win.addLib("javacc.zip","developer version only",true);
// 	libpath = win.addLib("mgtp.jar","developer version only",true);
// 	libpath = win.addLib("junit.jar","developer version only",true);
// 	libpath = win.addLib("cup.jar","developer version only",true);

     }

    // Inner static
    static class ProgressAdapter {

	private String action="";
	private double percent=0.0;
	

	public ProgressAdapter() {
	}

	public ProgressAdapter(String action, double percent) {
	    this.action=action;
	    this.percent=percent;
	}

	public void progressChanged(String action, double percent) {
	    setAction(action);
	    setPercent(percent);
	}

	public void setPercent(double d) {
	    this.percent=d;
	}

	public void setAction(String action) {
	    this.action=action;
	}

	public String getAction() {
	    return this.action;
	}

	public double getPercent() {
	    return this.percent;
	}
	
    }

    static class GUIProgressAdapter extends ProgressAdapter {

	private JFrame progressFrame;
	private JProgressBar progressBar;
	private JLabel label;

	public GUIProgressAdapter() {
	    this.progressFrame=new JFrame("Installing...");
	    this.progressBar=new JProgressBar(0, 100);
	    this.label=new JLabel(getAction());
	    progressBar.setStringPainted(true);	    
	    progressBar.setValue(0);
	    progressBar.setString(""+getPercent()*100);
	    progressFrame.getContentPane().setLayout(new GridLayout(2,1));
	    progressFrame.getContentPane().add(label);
	    progressFrame.getContentPane().add(progressBar);
	    progressFrame.pack();
	    progressFrame.setVisible(true);
	}

	public void setAction(String action) {
	    super.setAction(action);
	    label.setText(action);
	}
	
	public void setPercent(double d) {
	    super.setPercent(d);
	    progressBar.setValue((int)Math.round(d*100));
	    progressBar.setString(""+(int)Math.round(d*100)+"%");	    
	}	
    }
}
